from jolt import config
from jolt import log
from jolt.plugins import http


NAME = "artifactory"


class Artifactory(http.Http):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._uri = config.get(NAME, "uri")
        self._repository = config.get(NAME, "repository", "jolt")
        raise_error_if(not self._uri, "HTTP URI not configured")
        if self._uri[-1] != "/":
            self._uri += "/"
        self._upload = config.getboolean(NAME, "upload", True)
        self._download = config.getboolean(NAME, "download", True)
        self._disabled = False

    def _get_url(self, node, artifact):
        return "{uri}{repository}/{name}/{file}".format(
            uri=self._uri,
            repository=self._repository,
            name=node.name,
            file=fs.path.basename(artifact.get_archive_path()))

log.warning("[Artifactory] Deprecated plugin loaded. Use HTTP instead.")



import base64
import glob
import imp
import inspect
import os
import sys

from jolt.tasks import attributes
from jolt.tasks import Alias, Task, TaskGenerator, TaskRegistry, WorkspaceResource
from jolt.error import raise_error_if, raise_task_error_if
from jolt import config
from jolt import filesystem as fs
from jolt import log
from jolt import utils
from jolt.manifest import ManifestExtension
from jolt.manifest import ManifestExtensionRegistry


class Recipe(object):
    def __init__(self, path, joltdir=None, project=None, source=None):
        self.path = path
        self.basepath = os.path.basename(path)
        self.joltdir = joltdir
        self.project = project
        self.source = source
        self.tasks = []

    def load(self):
        raise_error_if(self.source is not None, "recipe already loaded: {}", self.path)

        with open(self.path) as f:
            self.source = f.read()

    def save(self):
        raise_error_if(self.source is None, "recipe source unknown: {}", self.path)

        with open(self.path, "w") as f:
            f.write(self.source)


class NativeRecipe(Recipe):
    @staticmethod
    def _is_abstract(cls):
        return cls.__dict__.get("abstract", False) or cls.__name__.startswith("_")

    @staticmethod
    def _is_generator(cls):
        return inspect.isclass(cls) and \
            issubclass(cls, TaskGenerator) and \
            not NativeRecipe._is_abstract(cls)

    @staticmethod
    def _is_task(cls):
        return inspect.isclass(cls) and \
            issubclass(cls, Task) and \
            not NativeRecipe._is_abstract(cls)

    def load(self):
        super(NativeRecipe, self).load()

        name = utils.canonical(self.path)
        module = imp.load_source("joltfile_{0}".format(name), self.path)

        tasks = inspect.getmembers(module, self._is_task)
        generators = inspect.getmembers(module, self._is_generator)

        for _, cls in generators:
            cls.joltdir = self.joltdir or os.path.dirname(self.path)
            classes = utils.as_list(cls().generate())
            tasks += [(c.__name__, c) for c in classes if self._is_task(c)]

        for name, task in tasks:
            task.name = task.name or task.__name__.lower()
            task.joltdir = self.joltdir or os.path.dirname(self.path)
            task.joltproject = self.project
            self.tasks.append(task)

        log.verbose("Loaded: {0}", self.path)


class Loader(object):
    def recipes(self):
        pass


class LoaderFactory(object):
    def create(self):
        raise NotImplementedError()


class NativeLoader(Loader):
    def __init__(self):
        self._recipes = []
        self._find_files()

    def _find_files(self):
        files = []
        oldpath = None
        curpath = os.getcwd()
        while not self._recipes and oldpath != curpath:
            files = glob.glob(fs.path.join(curpath, "*.jolt"))

            for filepath in files:
                recipe = NativeRecipe(filepath)
                self._recipes.append(recipe)

            oldpath = curpath
            curpath = os.path.dirname(oldpath)
        self.path = oldpath if self._recipes else None

    @property
    def recipes(self):
        return self._recipes


_loaders = []

def register(factory):
    raise_error_if(not issubclass(factory, LoaderFactory),
                   "{} is not a LoaderFactory", factory.__name__)
    _loaders.append(factory)


@register
class NativeLoaderFactory(LoaderFactory):
    def create(self):
        return NativeLoader()


@utils.Singleton
class JoltLoader(object):
    filename = "*.jolt"

    def __init__(self):
        self._recipes = []
        self._tasks = []
        self._path = None
        self._project_modules = {}
        self._project_recipes = {}
        self._project_resources = {}

    def _add_project_module(self, project, src):
        modules = self._project_modules.get(project, [])
        modules.append(src)
        self._project_modules[project] = modules

    def _get_project_modules(self, project):
        return self._project_modules.get(project, [])

    def _add_project_recipe(self, project, joltdir, src):
        recipes = self._project_recipes.get(project, [])
        recipes.append((joltdir, src))
        self._project_recipes[project] = recipes

    def _get_project_recipes(self, project):
        return self._project_recipes.get(project, [])

    def _add_project_resource(self, project, resource_name, resource_task):
        class ProjectResource(Alias):
            name =  project + "/" + resource_name
            requires = [resource_task]
        self._tasks.append(ProjectResource)
        resources = self._project_resources.get(project, [])
        resources.append((resource_name, resource_task))
        self._project_resources[project] = resources

    def _get_project_resources(self, project):
        return self._project_resources.get(project, [])

    def _load_project_recipes(self):
        for project, recipes in self._project_recipes.items():
            resources = [resource for _, resource in self._get_project_resources(project)]
            for joltdir, src in recipes:
                joltdir = fs.path.join(self.joltdir, joltdir) if joltdir else self.joltdir
                recipe = NativeRecipe(fs.path.join(self._path, src), joltdir, project)
                recipe.load()
                for task in recipe.tasks:
                    task._resources = resources
                    attributes.requires("_resources")(task)
                self._tasks += recipe.tasks

    def load(self, manifest=None):
        for factory in _loaders:
            loader = factory().create()
            for recipe in loader.recipes:
                recipe.load()
                self._recipes.append(recipe)
                self._tasks += recipe.tasks
            if len(loader.recipes) > 0:
                self.set_joltdir(loader.path)

        self._load_project_recipes()
        return self._tasks

    def load_plugin(self, filepath):
        plugin, ext = os.path.splitext(fs.path.basename(filepath))
        imp.load_source("jolt.plugins." + plugin, filepath)

    def load_plugins(self):
        searchpath = config.get("jolt", "pluginpath")
        searchpath = searchpath.split(":") if searchpath else []
        searchpath.append(fs.path.join(fs.path.dirname(__file__), "plugins"))

        import jolt.plugins

        for plugin in config.plugins():
            for path in searchpath:
                if "jolt.plugins." + plugin not in sys.modules:
                    module = fs.path.join(fs.path.dirname(__file__), path, plugin + ".py")
                    if fs.path.exists(module):
                        self.load_plugin(module)
                        continue

    @property
    def recipes(self):
        return self._recipes

    @property
    def tasks(self):
        return self._tasks

    @property
    def joltdir(self):
        return self._path

    def set_joltdir(self, value):
        if not self._path or len(value) < len(self._path):
            self._path = value


class RecipeExtension(ManifestExtension):
    def export_manifest(self, manifest, task):
        loader = JoltLoader.get()

        for recipe in loader.recipes:
            manifest_recipe = manifest.create_recipe()
            manifest_recipe.path = recipe.basepath
            manifest_recipe.source = recipe.source

        projects = set([task.task.joltproject for task in [task] + task.extensions + task.descendants])
        for project in filter(lambda x: x is not None, projects):
            manifest_project = manifest.create_project()
            manifest_project.name = project

            for name, resource_task in loader._get_project_resources(project):
                resource = manifest_project.create_resource()
                resource.name = name
                resource.text = resource_task

            for joltdir, src in loader._get_project_recipes(project):
                recipe = manifest_project.create_recipe()
                recipe.src = src
                if joltdir:
                    recipe.joltdir = joltdir

            for src in loader._get_project_modules(project):
                module = manifest_project.create_module()
                module.src = src


    def import_manifest(self, manifest):
        loader = JoltLoader.get()
        loader.set_joltdir(manifest.joltdir)

        for recipe in manifest.recipes:
            recipe = Recipe(recipe.path, source=recipe.source)
            recipe.save()

        for project in manifest.projects:
            for recipe in project.recipes:
                loader._add_project_recipe(project.name, recipe.joltdir, recipe.src)

            for resource in project.resources:
                loader._add_project_resource(project.name, resource.name, resource.text)

                # Acquire resource immediately
                task = TaskRegistry.get().get_task(resource.text, manifest=manifest)
                raise_task_error_if(not isinstance(task, WorkspaceResource), task,
                                    "only workspace resources are allowed in manifest")
                task.acquire_ws()

            for module in project.modules:
                loader._add_project_module(project.name, module.src)
                sys.path.append(fs.path.join(manifest.joltdir, module.src))

ManifestExtensionRegistry.add(RecipeExtension())




from jolt import Task
from jolt.error import raise_task_error_if


class CMake(Task):
    """ Builds and publishes a CMake project """

    abstract = True

    cmakelists = "CMakeLists.txt"
    """ Path to CMakeLists.txt or directory containing CMakelists.txt """

    options = []
    """ List of options and their values (``option[:type]=value``) """

    def run(self, deps, tools): 
        raise_task_error_if(not self.cmakelists, self, "cmakelists attribute has not been defined")

        cmake = tools.cmake()
        cmake.configure(tools.expand(self.cmakelists), *["-D" + tools.expand(option) for option in self.options])
        cmake.build()
        cmake.install()

    def publish(self, artifact, tools):
        cmake = tools.cmake()
        cmake.publish(artifact)



import click
import json
import os
import sys

from jolt import cache
from jolt import cli
from jolt import config
from jolt import filesystem as fs
from jolt import graph
from jolt import log
from jolt import scheduler
from jolt import utils
from jolt import loader
from jolt.error import raise_task_error_if
from jolt.hooks import TaskHook, TaskHookFactory, TaskHookRegistry
from jolt.influence import StringInfluence
from jolt.options import JoltOptions
from jolt.plugins import ninja
from jolt.tasks import TaskRegistry
from jolt.tools import Tools

log.verbose("[NinjaCompDB] Loaded")


def joltdir():
    return loader.JoltLoader.get().joltdir


def patch(command, attrib, search, replace):
    command[attrib] = command[attrib].replace(search, replace)


class CompDB(object):
    def __init__(self, path="compile_commands.json", artifact=None):
        self.commands = []
        self.attribs = {}
        if artifact:
            self.path = fs.path.join(artifact.path, path)
        else:
            self.path = path

    def metapath(self, path):
        return fs.path.join(fs.path.dirname(path), ".meta.compile_commands.json")

    def read(self, path=None):
        try:
            with open(path or self.path) as f:
                self.commands = json.load(f)
            try:
                with open(self.metapath(path or self.path)) as f:
                    self.attribs = json.load(f)
            except:
                self.attribs = {}
        except OSError:
            pass

    def write(self, path=None, force=False):
        if not force and not self.commands:
            return
        with open(path or self.path, "w") as f:
            json.dump(self.commands, f, indent=2)
        with open(self.metapath(path or self.path), "w") as f:
            json.dump(self.attribs, f, indent=2)

    def annotate(self, task):
        self.attribs["joltdir"] = joltdir()
        self.attribs["cachedir"] = config.get_cachedir()

    def relocate(self, task, sandboxes=False):
        for command in self.commands:
            utils.call_and_catch(patch, command, "command", self.attribs.get("joltdir", joltdir()), joltdir())
            utils.call_and_catch(patch, command, "command", self.attribs.get("cachedir", config.get_cachedir()), config.get_cachedir())
            if sandboxes:
                utils.call_and_catch(patch, command, "command", "sandbox-", "sandbox-reflect-")
            utils.call_and_catch(patch, command, "directory", self.attribs.get("joltdir", joltdir()), joltdir())

    def merge(self, db):
        self.commands.extend(db.commands)

def has_incpaths(artifact):
    return artifact.cxxinfo.incpaths.count() > 0

def stage_artifacts(artifacts, tools):
    for artifact in filter(has_incpaths, artifacts):
        tools.sandbox(artifact, incremental=True, reflect=fs.has_symlinks())

def get_task_artifacts(task, artifact=None):
    acache = cache.ArtifactCache.get()
    artifact = artifact or acache.get_artifact(task)
    return artifact, [acache.get_artifact(dep) for dep in task.children]


class CompDBHooks(TaskHook):
    def task_created(self, task):
        task.task.influence.append(StringInfluence("NinjaCompDB: v3"))

    def task_postrun(self, task, deps, tools):
        if not isinstance(task.task, ninja.CXXProject):
            return
        with tools.cwd(task.task.outdir):
            utils.call_and_catch(tools.run, "ninja -f build.ninja -t compdb > compile_commands.json")

    def task_postpublish(self, task, artifact, tools):
        if isinstance(task.task, ninja.CXXProject):
            with tools.cwd(task.task.outdir):
                artifact.collect("*compile_commands.json")

        # Add information about the workspace and cachedir roots
        db = CompDB(artifact=artifact)
        db.read()
        db.annotate(task)
        db.write()

        if isinstance(task.task, ninja.CXXProject):
            dbpath = fs.path.join(task.task.outdir, "all_compile_commands.json")
            _, deps = get_task_artifacts(task, artifact)
            db = CompDB(dbpath)
            for dep in [artifact] + deps:
                depdb = CompDB(artifact=dep)
                depdb.read()
                depdb.relocate(task)
                db.merge(depdb)
            db.write()
            artifact.collect(dbpath, flatten=True)

    def task_finished(self, task):
        if task.options.network or task.options.worker:
            return
        if isinstance(task.task, ninja.CXXProject):
            artifact, deps = get_task_artifacts(task)
            db = CompDB("all_compile_commands.json", artifact)
            db.read()
            db.relocate(task)
            outdir = task.tools.builddir("compdb", incremental=True)
            dbpath = fs.path.join(outdir, "all_compile_commands.json")
            db.write(dbpath, force=True)
            stage_artifacts(deps+[artifact], task.tools)

    def task_skipped(self, task):
        self.task_finished(task)


@TaskHookFactory.register
class CompDBHookFactory(TaskHookFactory):
    def create(self, env):
        return CompDBHooks()


@cli.cli.command(name="compdb")
@click.argument("task", type=str, nargs=-1, required=False, autocompletion=cli._autocomplete_tasks)
@click.option("-d", "--default", type=str, multiple=True, help="Override default parameter values.")
@click.pass_context
def compdb(ctx, task, default):
    """
    Generate a compilation database for a task.

    Aggregates compilation databases found in artifacts of the specified task and
    its dependencies. The commands are then post-processed and localized to the
    current workspace.

    All task artifacts are sandboxed and their directory trees are recreated
    using symlinks pointing to the origin of collected files. When opening a
    file, an IDE can then follow the symlinks into the workspace instead of
    opening files in the artifact cache.

    The database must be regenerated if dependencies or the directory tree
    of an artifact change.

    """

    manifest = ctx.obj["manifest"]
    options = JoltOptions(default=default)
    acache = cache.ArtifactCache.get(options)
    hooks = TaskHookRegistry.get(options)
    executors = scheduler.ExecutorRegistry.get(options)
    registry = TaskRegistry.get()
    strategy = scheduler.DownloadStrategy(executors, acache)
    queue = scheduler.TaskQueue(strategy)

    for params in default:
        registry.set_default_parameters(params)

    gb = graph.GraphBuilder(registry, manifest, options, progress=True)
    dag = gb.build(task)

    try:
        with log.progress("Progress", dag.number_of_tasks(), " tasks", estimates=False, debug=False) as p:
            while dag.has_tasks():
                leafs = dag.select(lambda graph, task: task.is_ready())

                # Order the tasks by their weights to improve build times
                leafs.sort(key=lambda x: x.weight)

                while leafs:
                    task = leafs.pop()
                    queue.submit(acache, task)

                task, error = queue.wait()
                p.update(1)

    except KeyboardInterrupt:
        print()
        log.warning("Interrupted by user")
        try:
            queue.abort()
            sys.exit(1)
        except KeyboardInterrupt:
            print()
            log.warning("Interrupted again, exiting")
            os._exit(1)

    for goal in dag.goals:
        artifact, deps = get_task_artifacts(goal)
        db = CompDB("all_compile_commands.json", artifact)
        db.read()
        db.relocate(goal, sandboxes=True)
        outdir = goal.tools.builddir("compdb", incremental=True)
        dbpath = fs.path.join(outdir, "all_compile_commands.json")
        db.write(dbpath, force=True)
        stage_artifacts(deps+[artifact], goal.tools)
        log.info("Compilation DB: {}", dbpath)



from contextlib import contextmanager
from datetime import datetime

from jolt import *
from jolt import config
from jolt import filesystem as fs
from jolt import log
from jolt import utils
from jolt.error import raise_error_if
from jolt.hooks import TaskHook, TaskHookFactory
from jolt.manifest import JoltManifest


class ReportHooks(TaskHook):
    def __init__(self):
        self.manifest = JoltManifest()

    def finalize_report(self, report, task, result):
        report.name = task.qualified_name
        report.duration = str(task.duration_running.seconds)
        report.goal = str(task.is_goal()).lower()
        report.identity = task.identity
        report.result = result
        if hasattr(task, "logstash"):
            report.logstash = task.logstash
        self.manifest.append(report)

    @contextmanager
    def task_run(self, task):
        try:
            yield
        except Exception as e:
            with task.task.report() as report:
                if not report.errors:
                    report.add_exception(e)
                self.finalize_report(report.manifest, task, "FAILED")
            raise e
        else:
            with task.task.report() as report:
                self.finalize_report(report.manifest, task, "SUCCESS")

    def write(self, filename):
        self.manifest.write(filename)

    @contextmanager
    def update(self):
        yield self.manifest

_report_hooks = ReportHooks()

def update():
    return _report_hooks.update()

def write(filename):
    return _report_hooks.write(filename)


@TaskHookFactory.register_with_prio(-10)
class ReportFactory(TaskHookFactory):
    def create(self, env):
        return _report_hooks



from base64 import decodebytes as base64_decodebytes
from base64 import encodebytes as base64_encodebytes
import codecs
from xml.etree.ElementTree import Element
from xml.etree.ElementTree import ElementTree
from xml.etree import ElementTree as ET


class SubElement(object):
    def __init__(self, tag='', elem=None):
        super(SubElement, self).__init__()
        self._elem = elem if elem is not None else Element(tag)

    def append(self, e, *args, **kwargs):
        return self._elem.append(e._elem, *args, **kwargs)

    @property
    def attrib(self):
        return self._elem.attrib

    @property
    def tag(self):
        return self._elem.tag

    @property
    def text(self):
        return self._elem.text

    @text.setter
    def text(self, value):
        self._elem.text = value

    @property
    def items(self):
        return self._elem.items

    @property
    def tail(self):
        return self._elem.tail

    def __len__(self):
        return len(self._elem)

    def get(self, *args, **kwargs):
        return self._elem.get(*args, **kwargs)

    def set(self, *args, **kwargs):
        return self._elem.set(*args, **kwargs)

    def iter(self, *args, **kwargs):
        return self._elem.iter(*args, **kwargs)


class Attribute(object):
    def __init__(self, attribute, varname=None, child=False, values=None, base64=False, zlib=False):
        self.attribute = attribute
        self.varname = varname if varname is not None else attribute.lower()
        self.child = child
        self.values = values
        self.base64 = base64 or zlib
        self.zlib = zlib

    def __call__(self, cls):
        def decorate(cls, attribute, varname, child, values):
            base64 = self.base64
            zlib = self.zlib

            def _check_value(value, values):
                if values and value not in values:
                    raise ValueError('{0} is not one of {1}'.format(value, values))

            def attr_get(self):
                if attribute not in self.attrib:
                    return ''
                value = self.get(attribute)
                if base64:
                    value = base64_decodebytes(value.encode())
                    if zlib:
                        value = codecs.decode(value, "zlib")
                    value = value.decode()
                return value

            def attr_set(self, value):
                if value is None:
                    try:
                        self.attrib.pop(attribute)
                    except:
                        pass
                    finally:
                        return
                _check_value(value, values)
                if base64:
                    value = value.encode()
                    if zlib:
                        value = codecs.encode(value, "zlib")
                    value = base64_encodebytes(value).decode()
                return self.set(attribute, value)

            def child_get(self):
                if not hasattr(self, '_'+varname):
                    e = SubElement(attribute, elem=self._elem.find(attribute))
                    setattr(self, '_'+varname, e)
                value = getattr(self, '_'+varname).text
                if value is None:
                    return None
                if base64:
                    try:
                        value = base64_decodebytes(value.encode())
                        if zlib:
                            value = codecs.decode(value, "zlib")
                        value = value.decode()
                    except:
                        value = getattr(self, '_'+varname).text
                return str(value)

            def child_set(self, value):
                _check_value(value, values)
                if value is None: return
                if not hasattr(self, '_'+varname):
                    e = SubElement(attribute)
                    self.append(e)
                    setattr(self, '_'+varname, e)
                if base64:
                    value = value.encode()
                    if zlib:
                        value = codecs.encode(value, "zlib")
                    value = base64_encodebytes(value).decode()
                getattr(self,'_'+varname).text = value

            if not child:
                setattr(cls, varname, property(attr_get, attr_set))
            else:
                setattr(cls, varname, property(child_get, child_set))
            return cls
        return decorate(cls, self.attribute, self.varname, self.child, self.values)


class Composition(object):
    def __init__(self, cls, name):
        self.cls = cls
        self.name = name if name is not None else cls.__name__.lower()

    def __call__(self, cls):
        def decorate(cls, comp_cls, name):
            def create(self, *args, **kwargs):
                child = comp_cls(*args, **kwargs)
                self.append(child)
                return child

            def remove(self, child, *args, **kwargs):
                self.remove(child._elem)

            @property
            def get(self):
                children = list(self.getroot()) if isinstance(self, ElementTree) else list(self._elem)
                children = [n for n in children if n.tag == name]
                return [comp_cls(elem=child) for child in children]

            setattr(cls, 'create_' + name, create)
            setattr(cls, 'remove_' + name, remove)
            setattr(cls, name + 's', get)
            return cls
        return decorate(cls, self.cls, self.name)



from jolt import *
from jolt.error import raise_task_error_if
from jolt.tasks import TaskRegistry
from jolt import utils

from functools import partial
from os import path
from platform import system


class DockerCLI(Download):
    """ Downloads and publishes the Docker command line client """
    name = "docker/cli"

    version = Parameter("20.10.9", help="Docker version")
    host = Parameter(system().lower(), help="Host operating system")
    arch = Parameter("x86_64", help="Host architecture")
    url = "https://download.docker.com/{host}/static/stable/{arch}/docker-{version}.tgz"

    def publish(self, artifact, tools):
        with tools.cwd(self._builddir):
            artifact.collect("docker/docker")
        artifact.environ.PATH.append("docker")


class DockerLogin(Resource):
    """
    Resource which logs in and out of a Docker Registry.

    If the user and password parameters are unset, credentials
    are fetched from the environment variables:

        - DOCKER_USER
        - DOCKER_PASSWD

    """
    name = "docker/login"

    requires = ["docker/cli"]

    user = Parameter("", help="Docker Registry username")
    passwd = Parameter("", help="Docker Registry password")

    def _user(self, tools):
        return str(self.user) or tools.getenv("DOCKER_USER")

    def _password(self, tools):
        return str(self.passwd) or tools.getenv("DOCKER_PASSWD")

    def acquire(self, artifact, deps, tools):
        raise_task_error_if(not self._user(tools), self, "Username has not been configured")
        raise_task_error_if(not self._password(tools), self, "Password has not been configured")

        with tools.cwd(tools.builddir()):
            tools.write_file("docker-credential", self._password(tools))
            tools.run("cat docker-credential | docker login -u {user} --password-stdin", user=self._user(tools))

    def release(self, artifact, deps, tools):
        tools.run("docker logout")


TaskRegistry.get().add_task_class(DockerCLI)
TaskRegistry.get().add_task_class(DockerLogin)


@influence.attribute("compression")
@influence.attribute("context")
@influence.attribute("dockerfile")
@influence.attribute("imagename")
@influence.attribute("tag")
class DockerImage(Task):
    """
    Builds a Docker image and publishes the resulting tarfile.

    The image may optionally be compressed using bzip2, gzip or lzma compression.

    Example:


    .. code-block:: python

        from jolt.plugins.docker import DockerImage

        class Busybox(DockerImage):
            compression = "gz"
            dockerfile = \"\"\"
            FROM busybox:latest
            CMD ["busybox"]
            \"\"\"
            tag = "busybox:latest"

    """
    abstract = True

    buildargs = []
    """
    List of build arguments and their values ("ARG=VALUE").

    The arguments are passed to Docker using --build-arg.
    """

    cleanup = True
    """ Remove image from Docker daemon upon completion. Default: True """

    compression = None
    """ Optional image compression "bz2", "gz", or "xz". """

    context = "."
    """ Path to build context, relative to joltdir (directory). """

    dockerfile = "Dockerfile"
    """ Path to the Dockerfile to build, or the full source code of such a file. """

    imagefile = "{canonical_name}.tar"
    """
    Name of the image tarball published by the task.

    If set to None, no image file will be saved and published.

    Defaults to the task's canonical name.
    """

    pull = True
    """
    Always pull images when building.

    Passes --pull to the Docker client.
    """

    push = False
    """
    Optionally push image to registry. Default: False

    To be able to push images, the current user must login to the Docker Registry.
    The ``docker/login`` Jolt resource can be used for that purpose.
    """

    tags = ["{canonical_name}:{identity}"]
    """ Optional list of image tags. Defaults to task's canonical name. """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def run(self, deps, tools):
        buildargs = " ".join(["--build-arg " + tools.expand(arg) for arg in self.buildargs])
        context = tools.expand_relpath(self.context, self.joltdir)
        dockerfile = tools.expand(self.dockerfile)
        self._imagefile = tools.expand(self.imagefile) if self.imagefile else None
        pull = " --pull" if self.pull else ""
        tags = [tools.expand(tag) for tag in self.tags]

        if not path.exists(dockerfile):
            with tools.cwd(tools.builddir()):
                tools.write_file("Dockerfile", self.dockerfile)
                dockerfile = tools.expand_relpath("Dockerfile", self.joltdir)
        else:
            dockerfile = tools.expand_path(self.dockerfile)
            dockerfile = tools.expand_relpath(dockerfile, context)

        self.info("Building image from {} in {}", dockerfile, context)
        with tools.cwd(context):
            image = tools.run("docker build . -f {} -t {} {}{}", dockerfile, tags[0], buildargs, pull)
            for tag in tags[1:]:
                tools.run("docker tag {} {}", tags[0], tag)

        try:
            if self.push:
                self.info("Pushing image")
                for tag in tags:
                    tools.run("docker push {}", tag)

            self.info("Saving image to file")
            with tools.cwd(tools.builddir()):
                if self._imagefile:
                    tools.run("docker image save {} -o {_imagefile}", tags[0])
                    if self.compression is not None:
                        tools.compress("{_imagefile}", "{_imagefile}.{compression}")
        finally:
            if self.cleanup:
                self.info("Removing image from Docker daemon")
                for tag in tags:
                    utils.call_and_catch(tools.run("docker image rm {}", tag))

    def publish(self, artifact, tools):
        artifact.strings.tag = tools.expand(self.tags[0])
        with tools.cwd(tools.builddir()):
            if self._imagefile:
                if self.compression is not None:
                    artifact.collect("{_imagefile}.{compression}")
                else:
                    artifact.collect("{_imagefile}")



import contextlib
import fnmatch
import re
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import partial
from threading import RLock
from string import Formatter
import os
import hashlib
import sys
from fasteners import lock, process_lock
import errno
import json


if sys.version_info[0] == 3:
    read_input = input
else:
    read_input = raw_input

try:
    from inspect import getattr_static
    getattr_safe = getattr_static
except:
    getattr_safe = getattr


locked = lock.locked


def is_str(s):
    try:
        return type(s) == str or type(s) == unicode
    except NameError:
        return type(s) == str

def decode_str(s):
    try:
        return s.decode()
    except:
        return s

def decorate_append(func, extra_func):
    def _f(*args, **kwargs):
        val = func(*args, **kwargs)
        extra_func(*args, **kwargs)
        return val
    return _f

def decorate_prepend(func, extra_func):
    def _f(*args, **kwargs):
        extra_func(*args, **kwargs)
        return func(*args, **kwargs)
    return _f

def as_list(t):
    return [t] if type(t) == str else list(t)

def as_stable_string_list(o):
    if type(o) == list or type(o) == tuple:
        return sorted([str(item) for item in o])
    elif type(o) == dict:
        return sorted(["{0}={1}".format(key, as_stable_string_list(val))
                       for key, val in o.items()])
    else:
        return [str(o)]

def as_stable_tuple_list(o):
    assert type(o) == dict, "as_stable_tuple_list: argument is not a dict"
    l = [(key, value) for key, value in o.items()]
    return sorted(l, key=lambda x: x[0])

def as_human_size(size):
    unit_precision = [("B", 0), ("KB", 0), ("MB", 1), ("GB", 2), ("TB", 2), ("PB", 2)]
    index = 0
    while size > 1024:
        size /= 1024
        index += 1
    return "{0} {1}".format(round(size, ndigits=unit_precision[index][1]), unit_precision[index][0])

def as_dirpath(dirpath):
    return dirpath + os.path.sep if dirpath[-1] != os.path.sep else dirpath

def call_or_return(obj, t):
    return t(obj) if callable(t) else t

def call_or_return_list(obj, t):
    return as_list(call_or_return(obj, t))

def call_and_catch(f, *args, **kwargs):
    try:
        return f(*args, **kwargs)
    except KeyboardInterrupt as e:
        raise e
    except:
        return None

def call_and_catch_and_log(f, *args, **kwargs):
    try:
        return f(*args, **kwargs)
    except KeyboardInterrupt as e:
        raise e
    except:
        from jolt import log
        log.exception()
        return None

def parse_task_name(name):
    if ":" in name:
        task, params = name.split(":", 1)
        params = params.split(",")
        def _param(param):
            if "=" in param:
                key, value = param.split("=", 1)
            else:
                key, value = param, None
            return key, value
        return task, {key: value for key, value in map(_param, params) if key}
    else:
        return name, {}

def format_task_name(name, params):
    if not params:
        return name
    def _param(key, value):
        return "{0}={1}".format(key, value) if value is not None else key
    params = sorted([(key, value) for key, value in params.items()], key=lambda x: x[0])
    return "{0}:{1}".format(name, ",".join([_param(key, value) for key, value in params]))

def stable_task_name(name):
    task, params = parse_task_name(name)
    return format_task_name(task, params)

def canonical(s):
    return "".join([c if c.isalnum() else '_' for c in s])


def unique_list(seq):
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


def pathmatch(string, pattern):
    pattern = fnmatch.translate(pattern)
    pattern = pattern.replace('(?s:.*.*/', '(?s:(^|.*/)')
    pattern = pattern.replace('/.*.*/', '.*/')
    return re.compile(pattern).match(string)


def quote_path(string):
    return '"' + string + '"' if ' ' in string else string


class _SafeDict(object):
    def __init__(self, values, ignore_errors=False):
        self.values = values
        self.errors = not ignore_errors

    def _envget(self, key):
        if key.startswith("ENV|"):
            return os.environ.get(key[4:])
        return None

    def __getitem__(self, key):
        value = self.values.get(key)
        if value is None:
            value = call_and_catch(getattr, self.values.get("_instance", object()), key)
        if value is None:
            value = self._envget(key)
        if value is not None:
            return value
        if self.errors:
            raise KeyError(key)
        return "{" + key + "}"


def expand(string, *args, **kwargs):
    ignore_errors = kwargs.get("ignore_errors") or False
    return Formatter().vformat(str(string), args, _SafeDict(kwargs, ignore_errors))


class duration(object):
    def __init__(self):
        self._time = time.time()

    def __str__(self):
        elapsed = self.seconds
        if elapsed >= 3600:
            return time.strftime("%Hh %Mmin %Ss", time.gmtime(elapsed))
        if elapsed >= 60:
            return time.strftime("%Mmin %Ss", time.gmtime(elapsed))
        return time.strftime("%Ss", time.gmtime(elapsed))

    def __le__(self, d):
        now = time.time()
        elapsed1 = now - self._time
        elapsed2 = now - d._time
        return elapsed1 < elapsed2

    def __sub__(self, delta):
        assert type(delta) in [int, float]
        self._time -= int(delta+.5)
        return self

    def diff(self, d):
        if d is None:
            return duration_diff(0)
        now = time.time()
        elapsed1 = now - self._time
        elapsed2 = now - d._time
        return duration_diff(abs(elapsed1 - elapsed2))

    @property
    def seconds(self):
        return time.time() - self._time

    @property
    def milliseconds(self):
        return (time.time() - self._time) * 1000

class duration_diff(object):
    def __init__(self, elapsed):
        self._elapsed = elapsed

    def __str__(self):
        elapsed = self._elapsed + 0.5
        if elapsed <= 1:
            return ""
        if elapsed >= 3600:
            return time.strftime("[%Hh %Mmin %Ss] ", time.gmtime(elapsed))
        if elapsed >= 60:
            return time.strftime("[%Mmin %Ss] ", time.gmtime(elapsed))
        return time.strftime("[%Ss] ", time.gmtime(elapsed))

    def __iadd__(self, dur):
        if isinstance(dur, duration):
            now = duration()
            self._elapsed += dur.diff(now)._elapsed
        if type(dur) == int:
            self._elapsed += dur
        return self

    @property
    def elapsed(self):
        return self._elapsed


class cached:
    mutex = RLock()

    @staticmethod
    def instance(f):
        def _f(self, *args, **kwargs):
            attr = "__cached_result_" + str(id(f))
            with cached.mutex:
                if not hasattr(self, attr):
                    setattr(self, attr, f(self, *args, **kwargs))
            return getattr(self, attr)
        return _f

    @staticmethod
    def method(f):
        def _f(*args, **kwargs):
            attr = "__cached_result_" + str(id(f))
            with cached.mutex:
                if not hasattr(f, attr):
                    setattr(f, attr, f(*args, **kwargs))
            return getattr(f, attr)
        return _f


class retried:
    @staticmethod
    def on_exception(exc_type, pattern=None, count=8, backoff=[1,4,10,15,20,25,35,40]):
        def _decorate(f):
            def _f(*args, **kwargs):
                for i in range(0, count):
                    try:
                        if i > 0:
                            time.sleep(backoff[i - 1])
                        return f(*args, **kwargs)
                    except exc_type as e:
                        if i+1 >= count:
                            raise e
                        if pattern is None or pattern in str(e):
                            from jolt import log
                            log.debug("Exception caught, retrying : " + str(e))
                            continue
            return _f
        return _decorate

def ignore_exception(exc=Exception):
    return contextlib.suppress(exc)

def Singleton(cls):
    cls._instance = None
    @staticmethod
    def get(*args, **kwargs):
        if not cls._instance:
            cls._instance = cls(*args, **kwargs)
        return cls._instance
    cls.get = get
    return cls


class LockFile(object):
    def __init__(self, path, logfunc=None, *args, **kwargs):
        self._file = process_lock.InterProcessLock(os.path.join(path, "lock"))
        if not self._file.acquire(blocking=False):
            if logfunc is not None:
                logfunc(*args, **kwargs)
            self._file.acquire()

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        pass


def map_consecutive(method, iterable):
    return list(map(method, iterable))

def map_concurrent(method, iterable, *args, **kwargs):
    max_workers = kwargs.get("max_workers", None)
    callables = [partial(method, item) for item in iterable]
    futures = []
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        for call in callables:
            futures.append(executor.submit(call))
        for future in as_completed(futures):
            try:
                future.result()
            except Exception:
                for future in futures:
                    future.cancel()
    return [future.result() for future in futures]


def sha1(string):
    sha = hashlib.sha1()
    sha.update(string.encode())
    return sha.hexdigest()


def filesha1(path):
    sha = hashlib.sha1()
    with open(path, "rb") as f:
        for data in iter(lambda: f.read(0x10000), b''):
            sha.update(data)
    return sha.hexdigest()


def fromjson(filepath, ignore_errors=False):
    try:
        with open(filepath) as f:
            return json.loads(f.read())
    except Exception as e:
        if ignore_errors:
            return {}
        raise e


def tojson(filepath, data, ignore_errors=False, indent=2):
    try:
        with open(filepath, "w") as f:
            f.write(json.dumps(data, indent=indent))
    except Exception as e:
        if ignore_errors:
            return
        raise e


def concat_attributes(attrib, postfix, prepend=False):
    def _decorate(cls):
        _orig = getattr(cls, "_" + attrib)
        def _get(self):
            orig = _orig(self)
            if prepend:
                return getattr(self, self.expand(postfix), type(orig)()) + orig
            else:
                return orig + getattr(self, self.expand(postfix), type(orig)())
        setattr(cls, "_" + attrib, _get)
        return cls
    return _decorate



from jolt.tasks import Task, TaskRegistry
from jolt.cache import ArtifactCache
from jolt.graph import GraphBuilder
from jolt.error import raise_error_if
from jolt.manifest import JoltManifest
from jolt.scheduler import JoltEnvironment
from jolt.scheduler import LocalExecutor
from jolt.scheduler import LocalExecutorFactory
from jolt.scheduler import NetworkExecutorExtension
from jolt.scheduler import NetworkExecutorExtensionFactory
from jolt.loader import JoltLoader
from jolt import config
from jolt import filesystem as fs
from jolt import influence
from jolt import log
from jolt import utils



log.verbose("[SelfDeploy] Loaded")

_path = fs.path.dirname(__file__)
_path = fs.path.dirname(_path)
_path = fs.path.dirname(_path)


@influence.files(fs.path.join(_path, "**", "*.py"))
@influence.files(fs.path.join(_path, "**", "*.xslt"))
class Jolt(Task):
    name = "jolt"

    def __init__(self, *args, **kwargs):
        super(Jolt, self).__init__(*args, **kwargs)
        self.influence.append(
            influence.StringInfluence(
                config.get("selfdeploy", "requires", "")))
        for e in self.extras:
            self.influence.append(influence.DirectoryInfluence(e))

    def _verify_influence(self, *args, **kwargs):
        pass  # FIXME: Validation doesn't work properly for directories

    @property
    def loaderdir(self):
        return JoltLoader.get().joltdir

    @property
    def extras(self):
        ext = config.get("selfdeploy", "extra", "")
        return [fs.path.join(self.loaderdir, e) for e in ext.split(",")] if ext else []

    def info(self, fmt, *args, **kwargs):
        log.verbose(fmt, *args, **kwargs)

    def publish(self, artifact, tools):
        with tools.cwd(_path):
            artifact.collect('README.rst')
            artifact.collect('setup.py')
            artifact.collect('jolt/*.py')
            artifact.collect('jolt/*.job')
            artifact.collect('jolt/*/*.py')
            artifact.collect('jolt/*/*/*.py')
            artifact.collect('jolt/*/*.xslt')
            artifact.collect('jolt/plugins/selfdeploy/README.rst', flatten=True)
            artifact.collect('jolt/plugins/selfdeploy/setup.py', flatten=True)
            for e in self.extras:
                with tools.cwd(fs.path.dirname(e)):
                    artifact.collect(fs.path.basename(e))


class SelfDeployExtension(NetworkExecutorExtension):
    @utils.cached.instance
    def get_parameters(self, task):
        registry = TaskRegistry()
        registry.add_task_class(Jolt)
        acache = ArtifactCache.get()
        env = JoltEnvironment(cache=acache)
        gb = GraphBuilder(registry, JoltManifest())
        dag = gb.build(["jolt"])
        task = dag.select(lambda graph, task: True)
        assert len(task) == 1, "too many selfdeploy tasks found"
        task = task[0]
        if not acache.is_available_remotely(task):
            factory = LocalExecutorFactory()
            executor = LocalExecutor(factory, task, force_upload=True)
            executor.run(env)
        jolt_url = acache.location(task)
        raise_error_if(not jolt_url, "failed to deploy jolt to a remote cache")
        return {
            "jolt_url":  jolt_url,
            "jolt_identity": task.identity[:8],
            "jolt_requires": config.get("selfdeploy", "requires", "") }


@NetworkExecutorExtensionFactory.Register
class SelfDeployExtensionFactory(NetworkExecutorExtensionFactory):
    def create(self):
        return SelfDeployExtension()



import inspect

from jolt import utils


class JoltError(Exception):
    def __init__(self, *args, **kwargs):
        super(JoltError, self).__init__(*args, **kwargs)


class JoltCommandError(JoltError):
    def __init__(self, what, stdout, stderr, returncode, *args, **kwargs):
        super(JoltCommandError, self).__init__(what, *args, **kwargs)
        self.stdout = stdout
        self.stderr = stderr
        self.returncode = returncode


def raise_error(msg, *args, **kwargs):
    raise JoltError(msg.format(*args, **kwargs))


def raise_task_error(task, msg, *args, **kwargs):
    if task:
        with utils.ignore_exception():
            with task.report() as report:
                report.add_error("Error", inspect.getsourcefile(task.__class__),
                                 msg.format(*args, **kwargs))
        raise_error(msg + " (" + str(task) + ")", *args, **kwargs)
    else:
        raise_error(msg, *args, **kwargs)


def raise_unreported_task_error(task, msg, *args, **kwargs):
    if task:
        raise_error(msg + " (" + str(task) + ")", *args, **kwargs)
    else:
        raise_error(msg, *args, **kwargs)


def raise_error_if(condition, *args, **kwargs):
    if condition:
        raise_error(*args, **kwargs)


def raise_task_error_if(condition, task, *args, **kwargs):
    if condition:
        raise_task_error(task, *args, **kwargs)


def raise_unreported_task_error_if(condition, task, *args, **kwargs):
    if condition:
        raise_unreported_task_error(task, *args, **kwargs)


class raise_error_on_exception(object):
    def __init__(self, message):
        self.message = message

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        raise_error_if(value, self.message)


class raise_task_error_on_exception(object):
    def __init__(self, task, *args, **kwargs):
        self.task = task
        self.args = args
        self.kwargs = kwargs

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        raise_task_error_if(value, self.task, *self.args, **self.kwargs)



from datetime import datetime

from jolt import *
from jolt import cache
from jolt import config
from jolt import filesystem as fs
from jolt import log
from jolt import utils
from jolt.error import raise_error_if
from jolt.hooks import TaskHook, TaskHookFactory


log.verbose("[Symlinks] Loaded")


class SymlinkHooks(TaskHook):
    def __init__(self):
        self._path = config.get("symlinks", "path", "artifacts")
        raise_error_if(not self._path, "symlinks.path not configured")

    def task_finished(self, task):
        if not task.has_artifact():
            return

        srcpath = cache.ArtifactCache.get().get_path(task)
        destpath = fs.path.join(
            task.task.joltdir,
            self._path,
            utils.canonical(task.short_qualified_name))

        if fs.path.exists(srcpath):
            fs.unlink(destpath, ignore_errors=True)
            fs.makedirs(fs.path.dirname(destpath))
            fs.symlink(srcpath, destpath)

    def task_pruned(self, task):
        self.task_finished(task)

    def task_skipped(self, task):
        self.task_finished(task)


@TaskHookFactory.register
class SymlinkFactory(TaskHookFactory):
    def create(self, env):
        return SymlinkHooks()



from concurrent.futures import ThreadPoolExecutor, as_completed, Future
from functools import partial
import os
import queue

from jolt import config
from jolt import hooks
from jolt import log
from jolt import utils
from jolt import tools
from jolt.error import raise_error
from jolt.error import raise_task_error
from jolt.error import raise_task_error_if
from jolt.graph import PruneStrategy
from jolt.manifest import ManifestExtension
from jolt.manifest import ManifestExtensionRegistry
from jolt.options import JoltOptions



class JoltEnvironment(object):
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class TaskQueue(object):
    def __init__(self, strategy):
        self.futures = {}
        self.strategy = strategy
        self.duration_acc = utils.duration_diff(0)
        self._aborted = False

    def submit(self, cache, task):
        if self._aborted:
            return None

        env = JoltEnvironment(cache=cache)
        executor = self.strategy.create_executor(task)
        raise_task_error_if(
            not executor, task,
            "no executor can execute the task; "
            "requesting a distributed network build without proper configuration?")

        task.set_in_progress()
        future = executor.submit(env)
        self.futures[future] = task
        return future

    def wait(self):
        for future in as_completed(self.futures):
            task = self.futures[future]
            try:
                future.result()
            except Exception as error:
                log.exception()
                return task, error
            finally:
                self.duration_acc += task.duration_running or 0
                del self.futures[future]
            return task, None
        return None, None

    def abort(self):
        self._aborted = True
        for future, task in self.futures.items():
            future.cancel()
        if len(self.futures):
            log.info("Waiting for tasks to finish, please be patient")
        self.strategy.executors.shutdown()

    def is_aborted(self):
        return self._aborted

    def in_progress(self, task):
        return task in self.futures.values()


class Executor(object):
    def __init__(self, factory):
        self.factory = factory

    def submit(self, env):
        return self.factory.submit(self, env)

    def is_aborted(self):
        return self.factory.is_aborted()

    def run(self, env):
        pass


class LocalExecutor(Executor):
    def __init__(self, factory, task, force_upload=False, force_build=False):
        super(LocalExecutor, self).__init__(factory)
        self.task = task
        self.force_build = force_build
        self.force_upload = force_upload

    def run(self, env):
        if self.is_aborted():
            return
        try:
            self.task.started()
            with hooks.task_run(self.task):
                self.task.run(
                    env.cache,
                    force_build=self.force_build,
                    force_upload=self.force_upload)
        except Exception as e:
            log.exception()
            self.task.failed()
            raise e
        else:
            self.task.finished()
        return self.task


class NetworkExecutor(Executor):
    pass


class SkipTask(Executor):
    def __init__(self, factory, task, *args, **kwargs):
        super(SkipTask, self).__init__(factory, *args, **kwargs)
        self.task = task

    def run(self, env):
        self.task.skipped()
        for ext in self.task.extensions:
            ext.skipped()
        return self.task


class Downloader(Executor):
    def __init__(self, factory, task, *args, **kwargs):
        super(Downloader, self).__init__(factory, *args, **kwargs)
        self.task = task

    def _download(self, env, task):
        if self.is_aborted():
            return
        if not task.is_downloadable():
            return
        try:
            task.started("Download")
            raise_task_error_if(
                not env.cache.download(task),
                task, "failed to download task artifact")
        except Exception as e:
            task.failed("Download")
            raise e
        else:
            task.finished("Download")

    def run(self, env):
        self._download(env, self.task)
        for ext in self.task.extensions:
            self._download(env, ext)
        return self.task


class Uploader(Executor):
    def __init__(self, factory, task, *args, **kwargs):
        super(Uploader, self).__init__(factory, *args, **kwargs)
        self.task = task

    def _upload(self, env, task):
        if self.is_aborted():
            return
        try:
            task.started("Upload")
            raise_task_error_if(
                not env.cache.upload(task),
                task, "failed to upload task artifact")
        except Exception as e:
            task.failed("Upload")
            raise e
        else:
            task.finished("Upload")

    def run(self, env):
        self._upload(env, self.task)
        for ext in self.task.extensions:
            self._upload(env, ext)

        return self.task


@utils.Singleton
class ExecutorRegistry(object):
    executor_factories = []
    extension_factories = []

    def __init__(self, options=None):
        self._options = options or JoltOptions()
        self._factories = [factory(self._options) for factory in self.__class__.executor_factories]
        self._local_factory = LocalExecutorFactory(self._options)
        self._concurrent_factory = ConcurrentLocalExecutorFactory(self._options)
        self._extensions = [factory().create() for factory in self.__class__.extension_factories]

    def shutdown(self):
        for factory in self._factories:
            factory.shutdown()
        self._local_factory.shutdown()
        self._concurrent_factory.shutdown()

    def create_skipper(self, task):
        return SkipTask(self._concurrent_factory, task)

    def create_downloader(self, task):
        # TODO: Switch to concurrent factory once the progress bar can handle it
        return Downloader(self._concurrent_factory, task)

    def create_uploader(self, task):
        # TODO: Switch to concurrent factory once the progress bar can handle it
        return Uploader(self._concurrent_factory, task)

    def create_local(self, task, force=False):
        task.set_locally_executed()
        return self._local_factory.create(task, force=force)

    def create_network(self, task):
        for factory in self._factories:
            executor = factory.create(task)
            if executor is not None:
                task.set_remotely_executed()
                return executor
        return self.create_local(task)

    def get_network_parameters(self, task):
        parameters = {}
        for extension in self._extensions:
            parameters.update(extension.get_parameters(task))
        return parameters


class NetworkExecutorExtensionFactory(object):
    @staticmethod
    def Register(cls):
        ExecutorRegistry.extension_factories.insert(0, cls)
        return cls

    def create(self):
        raise NotImplemented()


class NetworkExecutorExtension(object):
    def get_parameters(self, task):
        return {}


class Job(object):
    def __init__(self, priority, future, executor, env):
        self.priority = priority
        self.future = future
        self.executor = executor
        self.env = env

    def __le__(self, o):
        return self.priority <= o.priority

    def __ge__(self, o):
        return self.priority >= o.priority

    def __lt__(self, o):
        return self.priority < o.priority

    def __gt__(self, o):
        return self.priority > o.priority

    def __eq__(self, o):
        return self.priority == o.priority


class ExecutorFactory(object):
    @staticmethod
    def Register(cls):
        ExecutorRegistry.executor_factories.insert(0, cls)
        return cls

    def __init__(self, options=None, max_workers=None):
        self.pool = ThreadPoolExecutor(max_workers=max_workers)
        self._aborted = False
        self._queue = queue.PriorityQueue()
        self._options = options or JoltOptions()

    def is_aborted(self):
        return self._aborted

    def is_keep_going(self):
        return self._options.keep_going

    def shutdown(self):
        self._aborted = True
        self.pool.shutdown()

    def create(self, task):
        raise NotImplemented()

    def _run(self):
        job = self._queue.get(False)
        self._queue.task_done()
        try:
            if not self.is_aborted():
                job.executor.run(job.env)
        except KeyboardInterrupt as e:
            raise_error("Interrupted by user")
            self._aborted = True
            job.future.set_exception(e)
        except Exception as e:
            if not self.is_keep_going():
                self._aborted = True
            job.future.set_exception(e)
        else:
            job.future.set_result(job.executor)

    def submit(self, executor, env):
        future = Future()
        self._queue.put(Job(-executor.task.weight, future, executor, env))
        self.pool.submit(self._run)
        return future


class LocalExecutorFactory(ExecutorFactory):
    def __init__(self, options=None):
        max_workers = config.getint(
            "jolt", "parallel_tasks",
            os.getenv("JOLT_PARALLEL_TASKS", 1 if options is None else options.jobs))
        super(LocalExecutorFactory, self).__init__(
            options=options,
            max_workers=max_workers)

    def create(self, task, force=False):
        return LocalExecutor(self, task, force_build=force)


class ConcurrentLocalExecutorFactory(ExecutorFactory):
    def __init__(self, options=None):
        max_workers = tools.Tools().thread_count()
        super(ConcurrentLocalExecutorFactory, self).__init__(
            options=options,
            max_workers=max_workers)

    def create(self, task):
        raise NotImplementedError()


class NetworkExecutorFactory(ExecutorFactory):
    def __init__(self, *args, **kwargs):
        super(NetworkExecutorFactory, self).__init__(*args, **kwargs)


class ExecutionStrategy(object):
    def create_executor(self, task):
        raise NotImplemented()


class LocalStrategy(ExecutionStrategy, PruneStrategy):
    def __init__(self, executors, cache):
        self.executors = executors
        self.cache = cache

    def create_executor(self, task):
        if task.is_alias():
            return self.executors.create_skipper(task)
        if not task.is_cacheable():
            return self.executors.create_local(task)
        if task.is_available_locally(self.cache):
            return self.executors.create_skipper(task)
        if self.cache.download_enabled() and task.is_available_remotely(self.cache):
            return self.executors.create_downloader(task)
        return self.executors.create_local(task)

    def should_prune_requirements(self, task):
        if task.is_alias() or not task.is_cacheable():
            return False
        if task.is_available_locally(self.cache):
            return True
        if self.cache.download_enabled() and task.is_available_remotely(self.cache):
            return True
        return False


class DownloadStrategy(ExecutionStrategy, PruneStrategy):
    def __init__(self, executors, cache):
        self.executors = executors
        self.cache = cache

    def create_executor(self, task):
        if task.is_alias():
            return self.executors.create_skipper(task)
        if task.is_resource():
            return self.executors.create_local(task)
        if not task.is_cacheable():
            return self.executors.create_skipper(task)
        if task.is_available_locally(self.cache):
            return self.executors.create_skipper(task)
        if self.cache.download_enabled() and task.is_available_remotely(self.cache):
            return self.executors.create_downloader(task)
        raise_task_error(task, "task must be built first")

    def should_prune_requirements(self, task):
        return False


class DistributedStrategy(ExecutionStrategy, PruneStrategy):
    def __init__(self, executors, cache):
        self.executors = executors
        self.cache = cache

    def create_executor(self, task):
        if task.is_alias():
            return self.executors.create_skipper(task)

        if task.is_resource():
            if task.deps_available_locally(self.cache):
                return self.executors.create_local(task)
            else:
                return self.executors.create_skipper(task)

        if not task.is_cacheable():
            return self.executors.create_network(task)

        if not self.cache.upload_enabled():
            return self.executors.create_network(task)

        if not task.is_goal(with_extensions=False):
            task.disable_download()
        for extension in task.extensions:
            if not extension.is_goal(with_extensions=False):
                extension.disable_download()

        remote = task.is_available_remotely(self.cache)
        if remote:
            if task.is_goal() and self.cache.download_enabled() and \
               not task.is_available_locally(self.cache):
                return self.executors.create_downloader(task)
            return self.executors.create_skipper(task)
        else:
            if task.is_available_locally(self.cache) and task.is_uploadable(self.cache):
                return self.executors.create_uploader(task)
            if task.is_fast() and task.deps_available_locally(self.cache):
                return self.executors.create_local(task)

        return self.executors.create_network(task)

    def should_prune_requirements(self, task):
        if task.is_alias() or not task.is_cacheable():
            return False
        if task.is_available_remotely(self.cache):
            return True
        return False


class WorkerStrategy(ExecutionStrategy, PruneStrategy):
    def __init__(self, executors, cache):
        self.executors = executors
        self.cache = cache

    def create_executor(self, task):
        if task.is_resource():
            return self.executors.create_local(task)

        if task.is_alias():
            return self.executors.create_skipper(task)

        raise_task_error_if(
            not self.cache.upload_enabled(), task,
            "artifact upload must be enabled for workers, fix configuration")

        if not task.is_cacheable():
            return self.executors.create_local(task)

        if task.is_available_locally(self.cache):
            if task.is_goal() and not task.is_available_remotely(self.cache):
                # Unpacked artifacts may become unpacked before we manage to upload.
                # To keep the implementation simple we take the easy road and rebuild
                # all artifacts that have not been unpacked, even if they are uploadable.
                if task.is_unpacked(self.cache) and task.is_uploadable(self.cache):
                    return self.executors.create_uploader(task)
                else:
                    return self.executors.create_local(task, force=True)
            return self.executors.create_skipper(task)

        if not self.cache.download_enabled():
            return self.executors.create_local(task)

        if task.is_available_remotely(self.cache):
            return self.executors.create_downloader(task)

        return self.executors.create_local(task)

    def should_prune_requirements(self, task):
        if task.is_alias() or not task.is_cacheable():
            return False
        if task.is_available_locally(self.cache):
            # Unpacked artifacts may become unpacked before we manage to upload.
            # To keep the implementation simple we take the easy road and rebuild
            # all artifacts that have not been unpacked, even if they are uploadable.
            if task.is_unpacked(self.cache) and task.is_uploadable(self.cache):
                return True
        if not task.is_goal() and task.task.selfsustained:
            return True
        return False


class TaskIdentityExtension(ManifestExtension):
    def export_manifest(self, manifest, task):
        for child in [task] + task.extensions + task.descendants:
            manifest_task = manifest.find_task(child.qualified_name)
            if manifest_task is None:
                manifest_task = manifest.create_task()
                manifest_task.name = child.qualified_name
            manifest_task.identity = child.identity

ManifestExtensionRegistry.add(TaskIdentityExtension())


class TaskExportExtension(ManifestExtension):
    def export_manifest(self, manifest, task):
        short_task_names = set()
        for child in [task] + task.extensions + task.descendants:
            manifest_task = manifest.find_task(child.qualified_name)
            if manifest_task is None:
                manifest_task = manifest.create_task()
                manifest_task.name = child.qualified_name
            for key, export in child.task._get_export_objects().items():
                attrib = manifest_task.create_attribute()
                attrib.name = key
                attrib.value = export.export(child.task)
            short_task_names.add(child.name)

        # Figure out if any task with an overridden default parameter
        # value was included in the manifest. If so, add info about it.
        default_task_names = set()
        for task in task.options.default:
            short_name, _ = utils.parse_task_name(task)
            if short_name in short_task_names:
                default_task_names.add(task)
        if default_task_names:
            build = manifest.create_build()
            for task in default_task_names:
                default = build.create_default()
                default.name = task

ManifestExtensionRegistry.add(TaskExportExtension())



import copy
import os
import pygit2
import re

from jolt.tasks import Resource, WorkspaceResource, Parameter, BooleanParameter, Export, TaskRegistry
from jolt.influence import HashInfluenceProvider, HashInfluenceRegistry, FileInfluence
from jolt.tools import Tools
from jolt.loader import JoltLoader
from jolt import config
from jolt import filesystem as fs
from jolt import log
from jolt import utils
from jolt.error import JoltCommandError
from jolt.error import raise_error
from jolt.error import raise_error_if
from jolt.error import raise_task_error_if


log.verbose("[Git] Loaded")


class GitRepository(object):
    def __init__(self, url, path, relpath, refspecs=None):
        self.path = path
        self.relpath = relpath
        self.tools = Tools()
        self.url = url
        self.refspecs = refspecs or []
        self._tree_hash = {}
        self._original_head = True
        self._init_repo()

    def _init_repo(self):
        self.repository = pygit2.Repository(self.path) if os.path.exists(self._git_folder()) else None

    @utils.cached.instance
    def _git_folder(self):
        return fs.path.join(self.path, ".git")

    @utils.cached.instance
    def _git_index(self):
        return fs.path.join(self.path, ".git", "index")

    @utils.cached.instance
    def _git_jolt_index(self):
        return fs.path.join(self.path, ".git", "jolt-index")

    def is_cloned(self):
        return fs.path.exists(self._git_folder())

    def is_indexed(self):
        return fs.path.exists(self._git_index())

    def clone(self):
        log.info("Cloning into {0}", self.path)
        if fs.path.exists(self.path):
            with self.tools.cwd(self.path):
                self.tools.run("git init && git remote add origin {} && git fetch",
                               self.url, output_on_error=True)
        else:
            self.tools.run("git clone {0} {1}", self.url, self.path, output_on_error=True)
        raise_error_if(
            not fs.path.exists(self._git_folder()),
            "git: failed to clone repository '{0}'", self.relpath)
        self._init_repo()

    @utils.cached.instance
    def diff_unchecked(self):
        if not self.is_indexed():
            return ""

        # Build the jolt index file
        self.write_tree()

        # Diff against the jolt index
        with self.tools.environ(GIT_INDEX_FILE=self._git_jolt_index()), self.tools.cwd(self.path):
            return self.tools.run("git diff --binary --no-ext-diff HEAD ./",
                                  output_on_error=True,
                                  output_rstrip=False)

    def diff(self):
        diff = self.diff_unchecked()
        dlim = config.getsize("git", "maxdiffsize", "1M")
        raise_error_if(
            len(diff) > dlim,
            "git patch for '{}' exceeds configured size limit of {} bytes - actual size {}"
            .format(self.path, dlim, len(diff)))
        return diff

    def patch(self, patch):
        if not patch:
            return
        with self.tools.cwd(self.path), self.tools.tmpdir("git") as t:
            patchfile = fs.path.join(t.path, "jolt.diff")
            with open(patchfile, "wb") as f:
                f.write(patch.encode())
            log.info("Applying patch to {0}", self.path)
            self.tools.run("git apply --whitespace=nowarn {patchfile}", patchfile=patchfile)

    def head(self):
        if not self.is_cloned():
            return None
        return str(self.repository.head.target)

    def is_head(self, rev):
        return self.is_indexed() and self.head() == rev

    def is_original_head(self):
        return self._original_head

    def is_valid_sha(self, rev):
        return re.match(r"[0-9a-f]{40}", rev)

    def rev_parse(self, rev):
        if self.is_valid_sha(rev):
            return rev
        with self.tools.cwd(self.path):
            try:
                commit = self.repository.revparse_single(rev)
            except KeyError:
                self.fetch()
                try:
                    commit = self.repository.revparse_single(rev)
                except Exception as e:
                    raise_error("invalid git reference: {}", rev)
            try:
                return str(commit.id)
            except:
                return str(commit)

    @utils.cached.instance
    def write_tree(self):
        tools = Tools()
        with tools.cwd(self._git_folder()):
            tools.copy("index", "jolt-index")
        with tools.environ(GIT_INDEX_FILE=self._git_jolt_index()), tools.cwd(self.path):
            tree = tools.run(
                "git -c core.safecrlf=false add -A && git write-tree",
                output_on_error=True)
        return tree

    def tree_hash(self, sha=None, path="/"):
        # When sha is None, the caller want the tree hash of the repository's
        # current workspace state. If no checkout has been made, that would be the
        # tree that was written upon initialization of the repository as it
        # includes any uncommitted changes. If a checkout has been made since
        # the repo was initialized, make this an explicit request for the current
        # head - there can be no local changes.
        if sha is None:
            if self.is_original_head():
                tree = self.repository.get(self.write_tree())
            else:
                sha = self.head()

        path = fs.path.normpath(path)
        full_path = fs.path.join(self.path, path) if path != "/" else self.path

        # Lookup tree hash value in cache
        value = self._tree_hash.get((full_path, sha))
        if value is not None:
            return value

        # Translate explicit sha to tree
        if sha is not None:
            commit = self.rev_parse(sha)
            obj = self.repository.get(commit)
            try:
                tree = obj.tree
            except AttributeError:
                tree = obj.get_object().tree

        # Traverse tree from root to requested path
        if path != "/":
            tree = tree[fs.as_posix(path)]

        # Update tree hash cache
        self._tree_hash[(full_path, sha)] = value = tree.id

        return value

    def clean(self):
        with self.tools.cwd(self.path):
            return self.tools.run("git clean -dfx", output_on_error=True)

    def reset(self):
        with self.tools.cwd(self.path):
            return self.tools.run("git reset --hard", output_on_error=True)

    def fetch(self):
        refspecs = self.refspecs or []
        for refspec in [''] + refspecs:
            with self.tools.cwd(self.path):
                log.info("Fetching {0} from {1}", refspec or 'commits', self.url)
                self.tools.run("git fetch {url} {refspec}",
                               url=self.url,
                               refspec=refspec or '',
                               output_on_error=True)

    def checkout(self, rev):
        log.info("Checking out {0} in {1}", rev, self.path)
        with self.tools.cwd(self.path):
            try:
                return self.tools.run("git checkout -f {rev}", rev=rev, output=False)
            except:
                self.fetch()
                return self.tools.run("git checkout -f {rev}", rev=rev, output_on_error=True)
        self._original_head = False



_gits = {}

def new_git(url, path, relpath, refspecs=None):
    refspecs = utils.as_list(refspecs or [])
    try:
        git = _gits[path]
        raise_error_if(git.url != url, "multiple git repositories required at {}", relpath)
        raise_error_if(git.refspecs != refspecs,
                       "conflicting refspecs detected for git repository at  {}", relpath)
        return git
    except:
        git = _gits[path] = GitRepository(url, path, relpath, refspecs)
        return git



class GitInfluenceProvider(FileInfluence):
    name = "Git"

    def __init__(self, path):
        super(GitInfluenceProvider, self).__init__(path)
        self.path = path.rstrip(fs.sep)
        self.name = GitInfluenceProvider.name

    @property
    def joltdir(self):
        return JoltLoader.get().joltdir

    def _find_dotgit(self, path):
        ppath = None
        while path != ppath:
            if fs.path.isdir(fs.path.join(path, ".git")):
                return path
            ppath = path
            path = fs.path.dirname(path)
        raise_error("no git repository found at '{}'", self.path)

    @utils.cached.instance
    def get_influence(self, task):
        tools = Tools(task, task.joltdir)
        path = tools.expand_path(self.path)
        git_abs = self._find_dotgit(path)
        git_rel = git_abs[len(self.joltdir)+1:]
        relpath = path[len(git_abs)+1:]
        relpath = fs.as_posix(relpath) if relpath else relpath

        if not fs.path.exists(path):
            return "{0}/{1}: N/A".format(git_rel, relpath)
        try:
            git = new_git(None, git_abs, git_rel)

            return "{0}/{1}: {2}".format(git_rel, relpath, git.tree_hash(path=relpath or "/"))
        except JoltCommandError as e:
            stderr = "\n".join(e.stderr)
            if "exists on disk, but not in" in stderr:
                return "{0}/{1}: N/A".format(git_rel, relpath)
            raise e

    def is_influenced_by(self, task, path):
        tools = Tools(task, task.joltdir)
        gitpath = tools.expand_path(self.path)
        return fs.is_relative_to(path, gitpath)


def global_influence(path, cls=GitInfluenceProvider):
    HashInfluenceRegistry.get().register(cls(path))


def influence(path, git_cls=GitInfluenceProvider):
    def _decorate(cls):
        _old_influence = cls._influence
        def _influence(self, *args, **kwargs):
            influence = _old_influence(self, *args, **kwargs)
            influence.append(git_cls(path=path))
            return influence
        cls._influence = _influence
        return cls
    return _decorate


class GitSrc(WorkspaceResource, FileInfluence):
    """ Clones a Git repo.
    """

    name = "git-src"
    url = Parameter(help="URL to the git repo to be cloned. Required.")
    sha = Parameter(required=False, help="Specific commit or tag to be checked out. Optional.")
    path = Parameter(required=False, help="Local path where the repository should be cloned.")
    defer = BooleanParameter(False, help="Defer cloning until a consumer task must be built.")
    _revision = Export(value=lambda t: t._export_revision())
    _diff = Export(value=lambda t: t.git.diff(), encoded=True)

    def __init__(self, *args, **kwargs):
        super(GitSrc, self).__init__(*args, **kwargs)
        self.joltdir = JoltLoader.get().joltdir
        self.relpath = str(self.path) or self._get_name()
        self.abspath = fs.path.join(self.joltdir, self.relpath)
        self.refspecs = kwargs.get("refspecs", [])
        self.git = new_git(self.url, self.abspath, self.relpath, self.refspecs)

    @utils.cached.instance
    def _get_name(self):
        repo = fs.path.basename(self.url.get_value())
        name, _ = fs.path.splitext(repo)
        return name

    def _export_revision(self):
        return self.sha.value or self.git.head()

    def _get_revision(self):
        if self._revision.is_imported:
            return self._revision.value
        if not self.sha.is_unset():
            return self.sha.get_value()
        return None

    def acquire(self, artifact, env, tools):
        self._acquire_ws()

    def acquire_ws(self):
        if self.defer is None or self.defer.is_false:
            self._acquire_ws()

    def _acquire_ws(self):
        if not self.git.is_cloned():
            self.git.clone()
        if not self._revision.is_imported:
            self.git.diff_unchecked()
        rev = self._get_revision()
        if rev is not None:
            raise_task_error_if(
                not self._revision.is_imported and not self.sha.is_unset() and self.git.diff(), self,
                "explicit sha requested but git repo '{0}' has local changes", self.git.relpath)
            # Should be safe to do this now
            rev = self.git.rev_parse(rev)
            if not self.git.is_head(rev) or self._revision.is_imported:
                self.git.checkout(rev)
                self.git.clean()
                self.git.patch(self._diff.value)

    def get_influence(self, task):
        return None

    def is_influenced_by(self, task, path):
        return fs.is_relative_to(path, self.abspath) and self.sha.is_set()


TaskRegistry.get().add_task_class(GitSrc)


class Git(GitSrc):
    """ Clones a Git repo.

    Also influences the hash of consuming tasks, causing tasks to
    be re-executed if the cloned repo is modified.

    """
    name = "git"
    url = Parameter(help="URL to the git repo to be cloned. Required.")
    sha = Parameter(required=False, help="Specific commit or tag to be checked out. Optional.")
    path = Parameter(required=False, help="Local path where the repository should be cloned.")
    defer = None
    _revision = Export(value=lambda t: t._export_revision())
    _diff = Export(value=lambda t: t.git.diff(), encoded=True)

    def __init__(self, *args, **kwargs):
        super(Git, self).__init__(*args, **kwargs)
        self.influence.append(self)

    @utils.cached.instance
    def get_influence(self, task):
        if not self.git.is_cloned():
            self.git.clone()
        if not self._revision.is_imported:
            self.git.diff_unchecked()
        rev = self._get_revision()
        if rev is not None:
            return self.git.tree_hash(rev)
        return "{0}: {1}".format(
            self.git.relpath,
            self.git.tree_hash())

    def is_influenced_by(self, task, path):
        return path.startswith(self.abspath + fs.sep)


TaskRegistry.get().add_task_class(Git)



import os
import uuid
import errno

from jolt import utils
from jolt import cache
from jolt import log
from jolt import config
from jolt import filesystem as fs
from jolt.error import raise_error_if


NAME = "volume"
TIMEOUT = (3.5, 27)


class StaleFileHandleError(OSError):
    pass


class DiskVolume(cache.StorageProvider):
    def __init__(self, cache):
        super(DiskVolume, self).__init__()
        self._cache = cache
        self._path = config.get(NAME, "path")
        raise_error_if(not self._path, "volume path not configured")
        fs.makedirs(self._path)
        self._upload = config.getboolean(NAME, "upload", True)
        self._download = config.getboolean(NAME, "download", True)

    def _get_path(self, node, artifact):
        return "{path}/{name}/{file}".format(
            path=self._path,
            name=node.name,
            file=fs.path.basename(artifact.get_archive_path()))

    def _get_temp(self, node, artifact):
        return "{path}/{name}/{file}".format(
            path=self._path,
            name=node.name,
            file=uuid.uuid4())

    @utils.retried.on_exception(StaleFileHandleError)
    def download(self, node, force=False):
        if not self._download and not force:
            return False

        with self._cache.get_artifact(node) as artifact:
            path = self._get_path(node, artifact)
            try:
                log.verbose("[VOLUME] Copying {}", path)
                fs.copy(path, artifact.get_archive_path())
                return True
            except OSError as e:
                if e.errno == errno.ESTALE:
                    log.verbose("[VOLUME] got stale file handle, retrying...")
                    raise StaleFileHandleError(e)
                else:
                    log.exception()
            except Exception as e:
                log.exception()

        return False

    def download_enabled(self):
        return self._download

    def upload(self, node, force=False):
        if not self._upload and not force:
            return True
        with self._cache.get_artifact(node) as artifact:
            path = self._get_path(node, artifact)
            temp = self._get_temp(node, artifact)
            try:
                log.verbose("[VOLUME] Copying {}", path)
                fs.copy(artifact.get_archive_path(), temp)
                # To avoid race-condition, make sure that the artifact still is missing before moving it into place.
                if not fs.exists(path):
                    fs.rename(temp, path)
                else:
                    fs.unlink(temp)
                return True
            except OSError as e:
                if e.errno != errno.EEXIST:
                    log.verbose("[VOLUME] Failed to copy artifact, errno={}", os.strerror(e.errno))
                return e.errno == errno.EEXIST
            except Exception as e:
                log.exception()
            finally:
                fs.unlink(temp, ignore_errors=True)
        return False

    def upload_enabled(self):
        return self._upload

    def location(self, node):
        with self._cache.get_artifact(node) as artifact:
            path = self._get_path(node, artifact)
            avail = fs.path.exists(path)
            log.debug("[VOLUME] {} is{} present", path, "" if avail else " not")
            return avail
        return False


@cache.RegisterStorage
class DiskVolumeFactory(cache.StorageProviderFactory):
    @staticmethod
    def create(cache):
        log.verbose("[Volume] Loaded")
        return DiskVolume(cache)

# vim: et sw=4 ts=4



from socket import gethostname
from requests.exceptions import RequestException

from jolt import config
from jolt import log
from jolt import utils
from jolt.error import raise_error_if
from jolt.hooks import TaskHook, TaskHookFactory


log.verbose("[Telemetry] Loaded")


class TelemetryHooks(TaskHook):
    def __init__(
            self,
            plugin="telemetry",
            uri=None,
            local=True,
            network=True,
            queued=True,
            started=True,
            failed=True,
            finished=True):
        self._uri = uri or config.get(plugin, "uri", uri)
        self._network = config.getboolean(plugin, "network", network)
        self._local = config.getboolean(plugin, "local", local)
        self._queued = config.getboolean(plugin, "queued", queued)
        self._started = config.getboolean(plugin, "started", started)
        self._failed = config.getboolean(plugin, "failed", failed)
        self._finished = config.getboolean(plugin, "finished", finished)
        raise_error_if(not self._uri, "telemetry.uri not configured")

    @utils.retried.on_exception((RequestException))
    def post(self, task, event, client):
        data = {
            "name": task.short_qualified_name,
            "identity": task.identity,
            "instance": task.task._instance.value,
            "hostname": gethostname(),
            "role": "client" if client else "worker",
            "event": event,
        }
        from requests.api import post
        r = post(self._uri, json=data)
        r.raise_for_status()

    def task_started(self, task):
        if task.is_remotely_executed():
            if self._network and self._queued:
                self.post(task, "queued", client=True)
        elif task.is_locally_executed():
            if self._local and not task.options.worker:
                if self._queued:
                    self.post(task, "queued", client=True)
                if self._started:
                    self.post(task, "started", client=True)
            if self._network and task.options.worker:
                if self._started:
                    self.post(task, "started", client=False)

    def task_failed(self, task):
        if not self._failed:
            return
        if task.is_locally_executed():
            if self._local and not task.options.worker:
                self.post(task, "failed", client=True)
            if self._network and task.options.worker:
                self.post(task, "failed", client=False)
        if task.is_remotely_executed():
            if self._network and self._failed:
                self.post(task, "failed", client=True)

    def task_finished(self, task):
        if not self._finished:
            return
        if task.is_locally_executed():
            if self._local and not task.options.worker:
                self.post(task, "finished", client=True)
            if self._network and task.options.worker:
                self.post(task, "finished", client=False)
        if task.is_remotely_executed():
            if self._network and self._finished:
                self.post(task, "finished", client=True)


@TaskHookFactory.register
class TelemetryFactory(TaskHookFactory):
    def create(self, env):
        if "telemetry" in config.plugins():
            return TelemetryHooks()
        return None



from contextlib import contextmanager
import copy
import hashlib
from os import getenv
from threading import RLock
from collections import OrderedDict
import uuid

from jolt.tasks import Alias, Resource, WorkspaceResource
#from jolt.utils import *
from jolt.influence import HashInfluenceRegistry, TaskRequirementInfluence
from jolt import log
from jolt import utils
from jolt import colors
from jolt import hooks
from jolt import tools
from jolt import filesystem as fs
from jolt.error import raise_error_if
from jolt.error import raise_task_error_if
from jolt.options import JoltOptions


class TaskProxy(object):
    def __init__(self, task, graph, options):
        self.task = task
        self.graph = graph
        self.options = options

        self.children = []
        self.ancestors = set()
        self.descendants = set()
        self.neighbors = []
        self.extensions = []
        self.duration_queued = None
        self.duration_running = None

        self._extended_task = None
        self._in_progress = False
        self._completed = False
        self._goal = False
        self._download = True
        self._local = False
        self._network = False
        hooks.task_created(self)

    def __hash__(self):
        return id(self)

    @property
    def tools(self):
        return self.task.tools

    @property
    def name(self):
        return self.task.name

    @property
    def canonical_name(self):
        return self.task.canonical_name

    @property
    def qualified_name(self):
        return self.task.qualified_name

    @property
    def short_qualified_name(self):
        return self.task.short_qualified_name

    @property
    def log_name(self):
        return "({0} {1})".format(self.short_qualified_name, self.identity[:8])

    @property
    def identity(self):
        if self.task.identity is not None:
            return self.task.identity

        # Acquire workspace resources before calculating the identity
        for c in self.children:
            if c.is_workspace_resource():
                c.task.acquire_ws()

        sha = hashlib.sha1()
        HashInfluenceRegistry.get().apply_all(self.task, sha)
        self.task.identity = sha.hexdigest()

        return str(self.task.identity)

    @identity.setter
    def identity(self, value):
        self.task.identity = value

    @property
    def instance(self):
        return self.task._instance.value

    @property
    def weight(self):
        return self.task.weight

    @weight.setter
    def weight(self, weight):
        self.task.weight =  weight

    def __str__(self):
        return "{0}{1}".format(self.short_qualified_name, "*" if self.is_extension() else '')

    def info(self, fmt, *args, **kwargs):
        self.task.info(fmt + " " + self.log_name, *args, **kwargs)

    def verbose(self, fmt, *args, **kwargs):
        log.verbose(fmt + " " + self.log_name, *args, **kwargs)

    def warning(self, fmt, *args, **kwargs):
        self.task.warning(fmt + " " + self.log_name, *args, **kwargs)

    def error(self, fmt, *args, **kwargs):
        self.task.error(fmt + " " + self.log_name, *args, **kwargs)

    def has_children(self):
        return len(self.children) > 0

    def has_ancestors(self):
        return len(self.ancestors) > 0

    def has_artifact(self):
        return self.is_cacheable() and not self.is_resource() and not self.is_alias()

    def has_extensions(self):
        return len(self.extensions) > 0

    def add_extension(self, task):
        if self.is_extension():
            self.get_extended_task().add_extension(task)
        self.extensions.append(task)

    def set_extended_task(self, task):
        self._extended_task = task

    def get_extended_task(self):
        if self.is_extension():
            return self._extended_task.get_extended_task()
        return self

    def deps_available_locally(self, cache):
        for c in self.children:
            if c.is_resource() or c.is_alias():
                continue
            if not c.is_available_locally(cache):
                return False
        return True

    def is_alias(self):
        return isinstance(self.task, Alias)

    def is_available_locally(self, cache):
        tasks = [self] + self.extensions
        return all(map(cache.is_available_locally, tasks))

    def is_available_remotely(self, cache):
        tasks = [self] + self.extensions
        return all(map(cache.is_available_remotely, tasks))

    def is_cacheable(self):
        return self.task.is_cacheable()

    def is_completed(self):
        return self._completed

    def is_downloadable(self):
        return self._download

    def is_extension(self):
        return self._extended_task is not None

    def is_fast(self):
        tasks = [self.task] + [e.task for e in self.extensions]
        return all([task.fast for task in tasks])

    def is_identified(self):
        return self.task.identity is not None

    def is_goal(self, with_extensions=True):
        return self._goal or (with_extensions and any([e.is_goal() for e in self.extensions]))

    def in_progress(self):
        return self._in_progress

    def is_ready(self):
        if self.in_progress():
            return False

        if self.is_extension():
            return False

        return self.graph.is_leaf(self)

    def is_locally_executed(self):
        return self._local

    def is_remotely_executed(self):
        return self._network

    def is_resource(self):
        return isinstance(self.task, Resource)

    def is_unpacked(self, cache):
        tasks = [self] + self.extensions
        return any(map(cache.is_unpacked, tasks))

    def is_uploadable(self, cache):
        tasks = [self] + self.extensions
        return all(map(cache.is_uploadable, tasks))

    def is_workspace_resource(self):
        return isinstance(self.task, WorkspaceResource)

    def disable_download(self):
        self._download = False

    def set_in_progress(self):
        self._in_progress = True

    def set_locally_executed(self):
        self._local = True

    def set_remotely_executed(self):
        self._network = True

    def set_goal(self):
        self._goal = True

    def finalize(self, dag, manifest):
        log.debug("Finalizing: " + self.short_qualified_name)
        self.manifest = manifest

        # Find all direct and transitive dependencies
        self.ancestors = set()
        self.descendants = set()

        self.neighbors = copy.copy(self.children)
        self.neighbors = sorted(self.neighbors, key=lambda n: n.qualified_name)

        for n in self.neighbors:
            self.descendants.add(n)
            self.descendants = self.descendants.union(n.descendants)
            if not n.task.selfsustained:
                self.children.extend(n.children)
            n.ancestors.add(self)

        # Exclude transitive alias and resources dependencies
        self.children = list(
            filter(lambda n: not n.is_alias() and (not n.is_resource() or \
                   dag.are_neighbors(self, n)),
                   utils.unique_list(self.children)))

        self.descendants = list(self.descendants)

        self.task.influence += [TaskRequirementInfluence(n) for n in self.neighbors]

        return self.identity

    def taint(self, salt=None):
        self.task.taint = salt or uuid.uuid4()
        if salt is None:
            # Only recalculate identity when build is forced, not when salted
            self.identity = None
            self.identity

    def started(self, what="Execution"):
        self.task.info(colors.blue(what + " started " + self.log_name))
        self.duration_queued = utils.duration()
        self.duration_running = utils.duration()
        hooks.task_started(self)

    def running(self, when=None):
        self.duration_running = utils.duration() if not when else when

    def failed(self, what="Execution"):
        self.error("{0} failed after {1} {2}", what,
                   self.duration_running,
                   self.duration_queued.diff(self.duration_running))
        self.graph.add_failed(self)
        hooks.task_failed(self)

    def passed(self, what="Execution"):
        hooks.task_passed(self)

    def finished(self, what="Execution"):
        raise_task_error_if(
            self.is_completed() and not self.is_extension(),
            self, "task has already been completed")
        self._completed = True
        try:
            self.graph.remove_node(self)
        except KeyError:
            self.warning("Pruned task was executed")
        self.task.info(colors.green(what + " finished after {0} {1}" + self.log_name),
                       self.duration_running,
                       self.duration_queued.diff(self.duration_running))
        hooks.task_finished(self)

    def skipped(self):
        self._completed = True
        try:
            self.graph.remove_node(self)
        except KeyError:
            pass
        hooks.task_skipped(self)

    def pruned(self):
        self._completed = True
        try:
            self.graph.remove_node(self)
        except KeyError:
            self.warning("Pruned task was already pruned")
        hooks.task_pruned(self)

    def clean(self, cache, expired):
        with self.tools:
            self.task.clean(self.tools)
            discarded = cache.discard(self, expired)
            if discarded:
                log.debug("Discarded: {} ({})", self.short_qualified_name, self.identity[:8])
            else:
                log.debug(" Retained: {} ({})", self.short_qualified_name, self.identity[:8])

    def run(self, cache, force_upload=False, force_build=False):
        with self.tools:
            tasks = [self] + self.extensions
            available_locally = available_remotely = False

            for child in self.children:
                if not child.has_artifact():
                    continue
                if not cache.is_available_locally(child):
                    raise_task_error_if(
                        not cache.download(child),
                        child, "failed to download task artifact")

            if not force_build:
                available_locally = all(map(cache.is_available_locally, tasks))
                if available_locally and not force_upload:
                    return
                available_remotely = cache.download_enabled() and \
                    all(map(cache.is_available_remotely, tasks))
                if not available_locally and available_remotely:
                    available_locally = cache.download(self)

            if force_build or not available_locally:
                with log.threadsink() as buildlog:
                    if self.task.is_runnable():
                        log.verbose("Host: {0}", getenv("HOSTNAME", "localhost"))

                    with cache.get_locked_artifact(self, discard=force_build) as artifact:
                        if not cache.is_available_locally(self) or self.has_extensions():
                            with cache.get_context(self) as context:
                                self.running()
                                with self.tools.cwd(self.task.joltdir):
                                    hooks.task_prerun(self, context, self.tools)
                                    if self.is_goal() and self.options.debug:
                                        log.info("Entering debug shell")
                                        self.task.shell(context, self.tools)
                                    self.task.run(context, self.tools)
                                    hooks.task_postrun(self, context, self.tools)

                                if not cache.is_available_locally(self):
                                    with self.tools.cwd(self.task.joltdir):
                                        hooks.task_prepublish(self, artifact, self.tools)
                                        self.task.publish(artifact, self.tools)
                                        self.task._verify_influence(context, artifact, self.tools)
                                        hooks.task_postpublish(self, artifact, self.tools)
                                    with open(fs.path.join(artifact.path, ".build.log"), "w") as f:
                                        f.write(buildlog.getvalue())
                                    cache.commit(artifact)
                                else:
                                    self.info("Publication skipped, already in local cache")
                        else:
                            self.info("Execution skipped, already in local cache")

                        # Must upload the artifact while still holding its lock, otherwise the
                        # artifact may become unpack():ed before we have a chance to.
                        if force_upload or force_build or not available_remotely:
                            raise_task_error_if(
                                not cache.upload(self, force=force_upload, locked=False) and cache.upload_enabled(),
                                self, "failed to upload task artifact")
            elif force_upload or not available_remotely:
                raise_task_error_if(
                    not cache.upload(self, force=force_upload) and cache.upload_enabled(),
                    self, "failed to upload task artifact")

            for extension in self.extensions:
                try:
                    extension.started()
                    with hooks.task_run(extension):
                        extension.run(cache, force_upload, force_build)
                except Exception as e:
                    extension.failed()
                    raise e
                else:
                    extension.finished()

    def report(self):
        return self.task.report()


class Graph(object):
    def __init__(self):
        self._mutex = RLock()
        self._failed = []
        self._children = OrderedDict()
        self._parents = OrderedDict()

    def add_node(self, node):
        with self._mutex:
            self._children[node] = OrderedDict()
            self._parents[node] = OrderedDict()

    def remove_node(self, node):
        with self._mutex:
            parents = self._parents[node].keys()
            for child in self._children[node]:
                del self._parents[child][node]
            for parent in parents:
                del self._children[parent][node]
            del self._children[node]
            del self._parents[node]

    def add_edges_from(self, edges):
        with self._mutex:
            for src, dst in edges:
                self._children[src][dst] = None
                self._parents[dst][src] = None

    def successors(self, node):
        return self._children[node].keys()

    def predecessors(self, node):
        return self._parents[node].keys()

    @property
    def topological_nodes(self):
        with self._mutex:
            G = self.clone()
            S = G.roots
            L = []
            while S:
                for n in S:
                    L.append(n)
                    G.remove_node(n)
                S = G.roots
            if len(G.nodes) > 0:
                log.debug("[GRAPH] Graph has cycles between these nodes:")
                for node in G.nodes:
                    log.debug("[GRAPH]   " + node.short_qualified_name)
                raise_error_if(len(G.nodes) > 0, "graph has cycles")
            return L

    def clone(self):
        g = Graph()
        g._children = OrderedDict()
        g._parents = OrderedDict()
        for k, v in self._children.items():
            g._children[k] = copy.copy(v)
        for k, v in self._parents.items():
            g._parents[k] = copy.copy(v)
        return g

    @property
    def nodes(self):
        with self._mutex:
            return self._children.keys()

    def number_of_tasks(self, filterfn=None):
        return len(list(filter(filterfn, self.nodes)) if filterfn else self.nodes)

    @property
    def failed(self):
        return self._failed

    def add_failed(self, task):
        self._failed.append(task)

    @property
    def tasks(self):
        with self._mutex:
            return [n for n in self.nodes]

    @property
    def roots(self):
        with self._mutex:
            return [n for n in self.nodes if self.is_root(n)]

    def has_tasks(self):
        with self._mutex:
            return len(self.nodes) > 0

    def get_task(self, qualified_name):
        with self._mutex:
            return self._nodes_by_name.get(qualified_name)

    def select(self, func):
        with self._mutex:
            return [n for n in self.nodes if func(self, n)]

    def debug(self):
        with self._mutex:
            log.debug("[GRAPH] Listing all nodes")
            for node in self.topological_nodes:
                log.debug("[GRAPH]   " + node.qualified_name + " ({})", len(self._children[node].keys()))

    def is_leaf(self, node):
        with self._mutex:
            return len(self._children[node].keys()) == 0

    def is_root(self, node):
        with self._mutex:
            return len(self._parents[node].keys()) == 0

    def is_orphan(self, node):
        with self._mutex:
            return self.is_root(node) and self.is_leaf(node)

    def are_neighbors(self, n1, n2):
        with self._mutex:
            return n2 in self._children[n1]


class GraphBuilder(object):
    def __init__(self, registry, manifest, options=None, progress=False):
        self.graph = Graph()
        self.nodes = {}
        self.registry = registry
        self.manifest = manifest
        self.progress = progress
        self.options = options or JoltOptions()

    def _get_node(self, progress, name):
        name = utils.stable_task_name(name)
        node = self.nodes.get(name)
        if not node:
            task = self.registry.get_task(name, manifest=self.manifest)
            node = self.nodes.get(task.qualified_name, None)
            if node is not None:
                return node
            node = TaskProxy(task, self.graph, self.options)
            self.nodes[node.short_qualified_name] = node
            self.nodes[node.qualified_name] = node
            if self.options.salt:
                node.taint(self.options.salt)
            self._build_node(progress, node)
            progress.update(1)
        return node

    def _build_node(self, progress, node):
        self.graph.add_node(node)

        if node.task.extends:
            extended_node = self._get_node(progress, node.task.extends)
            self.graph.add_edges_from([(node, extended_node)])
            node.set_extended_task(extended_node)
            extended_node.add_extension(node)
            node.children.append(extended_node)
            parent = extended_node.get_extended_task()
        else:
            parent = node

        for requirement in node.task.requires:
            child = self._get_node(progress, requirement)
            # Create direct edges from alias parents to alias children
            if child.is_alias():
                for child_child in child.children:
                    self.graph.add_edges_from([(parent, child_child)])
                    node.children.append(child_child)
            else:
                self.graph.add_edges_from([(parent, child)])
            node.children.append(child)

        node.children = utils.unique_list(node.children)

        return node

    @contextmanager
    def _progress(self, *args, **kwargs):
        if self.progress:
            with log.progress(*args, **kwargs) as p:
                yield p
        else:
            with log.progress_log(*args, **kwargs) as p:
                yield p

    def build(self, task_list, influence=True):
        with self._progress("Building graph", len(self.graph.tasks), "tasks") as progress:
            goals = [self._get_node(progress, task) for task in task_list]
            self.graph._nodes_by_name = self.nodes

        if influence:
            topological_nodes = self.graph.topological_nodes
            with self._progress("Collecting task influence", len(self.graph.tasks), "tasks") as p:
                for node in reversed(topological_nodes):
                    node.finalize(self.graph, self.manifest)
                    p.update(1)

            max_time = 0
            min_time = 0
            for node in topological_nodes:
                max_time += node.task.weight
                node.task.weight += max([a.weight for a in node.ancestors] + [0])
                min_time = max(node.task.weight, min_time)

            log.verbose("Time estimate: {}- {}",
                        utils.duration_diff(min_time),
                        utils.duration_diff(max_time))

        self.graph.requested_goals = goals
        self.graph.goals = []
        for goal in goals:
            goal.set_goal()
            self.graph.goals.append(goal)
            if goal.is_alias():
                for goal_alias in goal.neighbors:
                    goal_alias.set_goal()
                    self.graph.goals.append(goal_alias)

        return self.graph


class PruneStrategy(object):
    def should_prune_requirements(self, task):
        raise NotImplementedError()


class GraphPruner(object):
    def __init__(self, strategy):
        self.strategy = strategy
        self.retained = set()
        self.visited = set()

    def _check_node(self, node):
        if node in self.visited:
            return
        self.visited.add(node)
        self.retained.add(node)

        prune = self.strategy.should_prune_requirements(node)
        if not node.task.selfsustained or not prune or node.is_extension():
            utils.map_concurrent(self._check_node, node.neighbors)

    def prune(self, graph):
        for root in graph.roots:
            self._check_node(root)

        for node in graph.goals:
            self.retained.add(node)

        pruned = []
        for node in graph.nodes:
            if node not in self.retained:
                pruned.append(node)
            else:
                log.debug("Retained: {}", node.log_name)
                node.children = [c for c in node.children if c in self.retained]

        for node in pruned:
            log.debug("Excluded: {}", node.log_name)
            node.pruned()

        return graph



import sys

from colorama import Fore, Style
from jolt import config


_tty = sys.stdout.isatty() and sys.stderr.isatty()

def enabled():
    return _tty and config.getboolean("jolt", "colors", True)

def red(s):
    return Fore.RED + Style.BRIGHT + s + Style.RESET_ALL if enabled() else s

def yellow(s):
    return Fore.YELLOW + Style.BRIGHT + s + Style.RESET_ALL if enabled() else s

def green(s):
    return Fore.GREEN + Style.BRIGHT + s + Style.RESET_ALL if enabled() else s

def blue(s):
    return Fore.BLUE + Style.BRIGHT + s + Style.RESET_ALL if enabled() else s

def bright(s):
    return Style.BRIGHT + s + Style.RESET_ALL if enabled() else s



import click
import imp
import subprocess
import sys
import uuid
import webbrowser
from os import _exit, environ, getcwd

from jolt.tasks import TaskRegistry, Parameter
from jolt import scheduler
from jolt import graph
from jolt import cache
from jolt import colors
from jolt import filesystem as fs
from jolt import log
from jolt.log import logfile
from jolt import config
from jolt.loader import JoltLoader
from jolt import tools
from jolt import utils
from jolt.influence import HashInfluenceRegistry
from jolt.options import JoltOptions
from jolt import hooks
from jolt.manifest import JoltManifest
from jolt.error import JoltError
from jolt.error import raise_error
from jolt.error import raise_error_if
from jolt.error import raise_task_error_if
from jolt.plugins import report

debug_enabled = False
workdir = getcwd()


class ArgRequiredUnless(click.Argument):
    def __init__(self, *args, **kwargs):
        self.required_unless = kwargs.pop('required_unless')
        assert self.required_unless, "'required_unless' parameter required"
        super(ArgRequiredUnless, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        if self.required_unless not in opts:
            if not opts.get(self.name, None):
                self.required = True
        return super(ArgRequiredUnless, self).handle_parse_result(
            ctx, opts, args)


class PluginGroup(click.Group):
    def get_command(self, ctx, cmd_name):

        if ctx.params.get("verbose", False):
            log.set_level(log.VERBOSE)
        if ctx.params.get("extra_verbose", False):
            log.set_level(log.DEBUG)
        config_files = ctx.params.get("config_file") or []
        for config_file in config_files:
            log.verbose("Config: {0}", config_file)
            config.load_or_set(config_file)

        # Load configured plugins
        JoltLoader.get().load_plugins()

        return click.Group.get_command(self, ctx, cmd_name)


@click.group(cls=PluginGroup, invoke_without_command=True)
@click.option("-v", "--verbose", is_flag=True, help="Verbose output.")
@click.option("-vv", "--extra-verbose", is_flag=True, help="Extra verbose output.")
@click.option("-c", "--config", "config_file", multiple=True, type=str,
              help="Load a configuration file or set a configuration key.")
@click.option("-d", "--debug-exception", is_flag=True, hidden=True,
              help="Attach debugger on exception")
@click.option("-p", "--profile", is_flag=True, hidden=True,
              help="Profile code while running")
@click.option("-f", "--force", is_flag=True, default=False, hidden=True,
              help="Force rebuild of target tasks.")
@click.option("-s", "--salt", type=str, hidden=True,
              help="Add salt as task influence.")
@click.option("-g", "--debug", is_flag=True, default=False, hidden=True,
              help="Start debug shell before executing task.")
@click.option("-n", "--network", is_flag=True, default=False, hidden=True,
              help="Build on network.")
@click.option("-l", "--local", is_flag=True, default=False, hidden=True,
              help="Disable all network operations.")
@click.option("-k", "--keep-going", is_flag=True, default=False, hidden=True,
              help="Build as many tasks as possible, don't abort on first failure.")
@click.option("-j", "--jobs", type=int, default=1, hidden=True,
              help="Number of tasks allowed to execute in parallel (1). ")
@click.pass_context
def cli(ctx, verbose, extra_verbose, config_file, debug_exception, profile,
        force, salt, debug, network, local, keep_going, jobs):
    """
    Jolt - a task execution tool.

    When invoked without any commands and arguments, Jolt by default tries
    to execute and build the artifact of a task called `default`. If the
    default task doesn't exist this help text is printed. All other tasks
    can be executed with the build command. Note that most build command
    options can be used also when executing the default task without
    specifying a command. See the build command help for details.

    To execute the default task and its dependencies without stopping on
    failures, run:

      $ jolt -k

    """

    global debug_enabled
    debug_enabled = debug_exception

    log.verbose("Jolt host: {}", environ.get("HOSTNAME", "localhost"))
    log.verbose("Jolt install path: {}", fs.path.dirname(__file__))

    if ctx.invoked_subcommand in ["config"]:
        # Don't attempt to load any task recipes as they might require
        # plugins that are not yet configured.
        return

    if ctx.invoked_subcommand is None:
        build = ctx.command.get_command(ctx, "build")

    manifest = JoltManifest()
    utils.call_and_catch(manifest.parse)
    manifest.process_import()
    ctx.obj["manifest"] = manifest

    loader = JoltLoader.get()
    tasks = loader.load()
    for cls in tasks:
        TaskRegistry.get().add_task_class(cls)

    if ctx.invoked_subcommand in ["build", "clean"] and loader.joltdir:
        ctx.obj["workspace_lock"] = utils.LockFile(
            fs.path.join(loader.joltdir, "build"),
            log.info, "Workspace is locked by another process, please wait...")

    # If no command is given, we default to building the default task.
    # If the default task doesn't exist, help is printed inside build().
    if ctx.invoked_subcommand is None:
        task = config.get("jolt", "default", "default")
        taskname, _ = utils.parse_task_name(task)
        if TaskRegistry.get().get_task_class(taskname) is not None:
            ctx.invoke(build, task=[task], force=force, salt=salt, debug=debug,
                       network=network, local=local, keep_going=keep_going, jobs=jobs)
        else:
            print(cli.get_help(ctx))
            sys.exit(1)


def _autocomplete_tasks(ctx, args, incomplete):
    manifest = JoltManifest()
    utils.call_and_catch(manifest.parse)
    manifest.process_import()

    tasks = JoltLoader.get().load()
    tasks = [task.name for task in tasks if task.name.startswith(incomplete or '')]
    return sorted(tasks)


@cli.command()
@click.argument("task", type=str, nargs=-1, autocompletion=_autocomplete_tasks, cls=ArgRequiredUnless, required_unless="worker")
@click.option("-n", "--network", is_flag=True, default=False, help="Build on network.")
@click.option("-l", "--local", is_flag=True, default=False, help="Disable all network operations.")
@click.option("-k", "--keep-going", is_flag=True, default=False, help="Build as many tasks as possible, don't abort on first failure.")
@click.option("-i", "--identity", type=str, help="Expected hash identity")
@click.option("-j", "--jobs", type=int, default=1, help="Number of tasks allowed to execute in parallel (1). ")
@click.option("--no-download", is_flag=True, default=False,
              help="Don't download artifacts from remote storage")
@click.option("--no-upload", is_flag=True, default=False,
              help="Don't upload artifacts to remote storage")
@click.option("--download", is_flag=True, default=False,
              help="Do download artifacts from remote storage")
@click.option("--upload", is_flag=True, default=False,
              help="Do upload artifacts to remote storage")
@click.option("--worker", is_flag=True, default=False,
              help="Run with the worker build strategy", hidden=True)
@click.option("-d", "--default", type=str, multiple=True, help="Override default parameter values.")
@click.option("-f", "--force", is_flag=True, default=False, help="Force rebuild of target tasks.")
@click.option("-s", "--salt", type=str, help="Add salt as task influence.")
@click.option("-g", "--debug", is_flag=True, default=False,
              help="Start debug shell before executing task.")
@click.option("-c", "--copy", type=click.Path(),
              help="Copy artifact content to this directory upon completion.")
@click.option("--result", type=click.Path(), hidden=True,
              help="Write result manifest to this file.")
@click.pass_context
@hooks.cli_build
def build(ctx, task, network, keep_going, identity, default, local,
          no_download, no_upload, download, upload, worker, force,
          salt, copy, debug, result, jobs):
    """
    Execute a specific task.

    """
    raise_error_if(network and local,
                   "The -n and -l flags are mutually exclusive")

    raise_error_if(network and debug,
                   "The -g and -n flags are mutually exclusive")

    raise_error_if(no_download and download,
                   "The --download and --no-download flags are mutually exclusive")

    raise_error_if(no_upload and upload,
                   "The --upload and --no-upload flags are mutually exclusive")

    duration = utils.duration()

    task = list(task)
    task = [utils.stable_task_name(t) for t in task]

    if network:
        _download = config.getboolean("network", "download", True)
        _upload = config.getboolean("network", "upload", True)
    else:
        _download = config.getboolean("jolt", "download", True)
        _upload = config.getboolean("jolt", "upload", True)

    if local:
        _download = False
        _upload = False
    else:
        if no_download:
            _download = False
        if no_upload:
            _upload = False
        if download:
            _download = True
        if upload:
            _upload = True

    options = JoltOptions(network=network,
                          local=local,
                          download=_download,
                          upload=_upload,
                          keep_going=keep_going,
                          default=default,
                          worker=worker,
                          debug=debug,
                          salt=salt,
                          jobs=jobs)

    acache = cache.ArtifactCache.get(options)

    executors = scheduler.ExecutorRegistry.get(options)
    if worker:
        log.set_worker()
        log.verbose("Local build as a worker")
        strategy = scheduler.WorkerStrategy(executors, acache)
    elif network:
        log.verbose("Distributed build as a user")
        strategy = scheduler.DistributedStrategy(executors, acache)
    else:
        log.verbose("Local build as a user")
        strategy = scheduler.LocalStrategy(executors, acache)

    hooks.TaskHookRegistry.get(options)
    registry = TaskRegistry.get(options)

    for params in default:
        registry.set_default_parameters(params)

    manifest = ctx.obj["manifest"]

    for mb in manifest.builds:
        for mt in mb.tasks:
            task.append(mt.name)
        for mt in mb.defaults:
            registry.set_default_parameters(mt.name)

    if force:
        for goal in task:
            registry.get_task(goal, manifest=manifest).taint = uuid.uuid4()

    gb = graph.GraphBuilder(registry, manifest, options, progress=True)
    dag = gb.build(task)

    gp = graph.GraphPruner(strategy)
    dag = gp.prune(dag)

    goal_tasks = dag.goals
    goal_task_duration = 0

    queue = scheduler.TaskQueue(strategy)

    try:
        if not dag.has_tasks():
            return

        progress = log.progress(
            "Progress",
            dag.number_of_tasks(filterfn=lambda t: not t.is_resource()),
            " tasks",
            estimates=False,
            debug=debug)

        with progress:
            while dag.has_tasks():
                # Find all tasks ready to be executed
                leafs = dag.select(lambda graph, task: task.is_ready())

                # Order the tasks by their weights to improve build times
                leafs.sort(key=lambda x: x.weight)

                while leafs:
                    task = leafs.pop()
                    queue.submit(acache, task)

                task, error = queue.wait()

                if not task:
                    dag.debug()
                    break
                elif task.is_goal() and task.duration_running:
                    goal_task_duration += task.duration_running.seconds

                if not task.is_resource():
                    progress.update(1)

                if not keep_going and error is not None:
                    queue.abort()
                    raise error

        if dag.failed:
            log.error("List of failed tasks")
            for failed in dag.failed:
                log.error("- {}", failed.log_name.strip("()"))
            raise_error("no more tasks could be executed")

        for goal in goal_tasks:
            if acache.is_available_locally(goal):
                with acache.get_artifact(goal) as artifact:
                    log.info("Location: {0}", artifact.path)
                    if copy:
                        artifact.copy("*", utils.as_dirpath(fs.path.join(workdir, click.format_filename(copy))), symlinks=True)
    except KeyboardInterrupt:
        print()
        log.warning("Interrupted by user")
        try:
            queue.abort()
            sys.exit(1)
        except KeyboardInterrupt:
            print()
            log.warning("Interrupted again, exiting")
            _exit(1)
    finally:
        log.info("Total execution time: {0} {1}",
                 str(duration),
                 str(queue.duration_acc) if network else '')
        if result:
            with report.update() as manifest:
                manifest.duration = str(goal_task_duration)
                manifest.write(result)


@cli.command()
@click.argument("task", type=str, nargs=-1, required=False, autocompletion=_autocomplete_tasks)
@click.option("-d", "--deps", is_flag=True, help="Clean all task dependencies.")
@click.option("-e", "--expired", is_flag=True, help="Only clean expired tasks.")
@click.pass_context
def clean(ctx, task, deps, expired):
    """
    Removes task artifacts and intermediate files.

    When run without arguments, this command removes all task artifacts
    from the local cache, but no intermediate files are removed.

    When a task is specified, the task clean() method is invoked to remove
    any intermediate files still present in persistent build directories.
    Secondly, the task artifact will be removed from the local cache.
    Global caches are not affected. The --deps parameter can be used to also
    clean all dependencies of the specified task.

    By default, task artifacts are removed without considering any
    artifact expiration metadata. To only remove artifact which have expired,
    use the --expired parameter. Artifacts typically expire immediately after
    creation unless explicitly configured not to.
    """
    acache = cache.ArtifactCache.get()
    if task:
        task = [utils.stable_task_name(t) for t in task]
        registry = TaskRegistry.get()
        dag = graph.GraphBuilder(registry, ctx.obj["manifest"]).build(task)
        if deps:
            tasks = dag.tasks
        else:
            tasks = dag.goals
        for task in tasks:
            task.clean(acache, expired)
    else:
        acache.discard_all(expired)


@cli.command(name="config")
@click.option("-l", "--list", is_flag=True,
              help="List all configuration keys and values.")
@click.option("-d", "--delete", is_flag=True,
              help="Delete configuration key.")
@click.option("-g", "--global", "global_", is_flag=True,
              help="List, set or get configuration keys in the global config.")
@click.option("-u", "--user", is_flag=True,
              help="List, set or get configuration keys in the user config.")
@click.argument("key", type=str, nargs=1, required=False)
@click.argument("value", type=str, nargs=1, required=False)
@click.pass_context
def _config(ctx, list, delete, global_, user, key, value):
    """
    Configure Jolt.

    You can query/set/replace/unset configuration keys with this command.
    Key strings are constructed from the configuration section and the
    option separated by a dot.

    There are tree different configuration sources:

       - A global configuration file

       - A user configuration file

       - Temporary configuration passed on the command line.

    When reading, the values are read from all configuration sources.
    The options --global and --user can be used to tell the command to read
    from only one of the sources. If a configuration key is available from
    multiple sources, temporary CLI configuration has priority followed by
    the user configuration file and lastly the global configuration file.

    When writing, the new values are written to the user configuration by default.
    The options --global and --user can be used to tell the command to write
    to only one of the sources.

    When removing keys, the values are removed from all sources.
    The options --global and --user can be used to restrict removal to one of
    the sources.

    To assign a value to a key:

      $ jolt config jolt.default all   # Change name of the default task

    To list existing keys:

      $ jolt config -l                 # List all existing keys

      $ jolt config -l -g              # List keys in the global config file

      $ jolt config jolt.colors        # Display the value of a key.

    To delete an existing key:

      $ jolt config -d jolt.colors

    To pass temporary configuration:

      $ jolt -c jolt.colors=true config -l

    """

    if delete and not key:
        raise click.UsageError("--delete requires KEY")

    if not key and not list and not key:
        print(ctx.get_help())
        sys.exit(1)

    if global_ and user:
        raise click.UsageError("--global and --user are mutually exclusive")

    alias = None

    if global_:
        alias = "global"
    if user:
        alias = "user"

    def _print_key(section, opt):
        value = config.get(section, opt, alias=alias)
        raise_error_if(value is None, "no such key: {}".format(key))
        print("{} = {}".format(key, value))

    if list:
        for section, option, value in config.items(alias):
            if option:
                print("{}.{} = {}".format(section, option, value))
            else:
                print(section)
    elif delete:
        raise_error_if(config.delete(key, alias) <= 0,
                       "no such key: {}", key)
        config.save()
    elif key:
        section, opt = config.split(key)
        if value:
            raise_error_if(opt is None, "invalid configuration key: {}".format(key))
            config.set(section, opt, value, alias)
            try:
                config.save()
            except Exception as e:
                log.exception()
                raise_error("failed to write configuration file: {}".format(e))
        else:
            if opt:
                _print_key(section, opt)
            else:
                _print_section(section)


@cli.command()
@click.argument("task", type=str, nargs=-1, required=False, autocompletion=_autocomplete_tasks)
@click.option("-r", "--reverse", type=str, help="Display consumers of REVERSE if TASK is executed.")
@click.option("-c", "--cached", "show_cache", is_flag=True, help="Highlight cache presence with colors.")
@click.pass_context
def display(ctx, task, reverse=None, show_cache=False):
    """
    Display a task and its dependencies visually.

    """
    registry = TaskRegistry.get()
    gb = graph.GraphBuilder(registry, ctx.obj["manifest"])
    dag = gb.build(task, influence=show_cache)

    options = JoltOptions()
    acache = cache.ArtifactCache.get(options)


    if reverse:
        iterator = lambda task: list(dag.predecessors(task))
        reverse = utils.as_list(reverse)
        tasklist = dag.select(lambda graph, node: \
                             node.short_qualified_name in reverse or \
                             node.qualified_name in reverse)
    else:
        tasklist = dag.requested_goals
        iterator = lambda task: task.children


    if dag.has_tasks():
        def _display(task, indent=0, last=None):
            header = ""
            if indent > 0:
                for pipe in last[:-1]:
                    if pipe:
                        header += "\u2502 "
                    else:
                        header += "  "
                if last[-1]:
                    header += "\u251c\u2574"
                else:
                    header += "\u2514\u2574"

            if not show_cache:
                colorize = str
            elif task.is_cacheable() and not acache.is_available(task):
                colorize = colors.red
            else:
                colorize = colors.green

            print(header + colorize(task.short_qualified_name))
            children = iterator(task)
            for i in range(0, len(children)):
                _display(children[i], indent+1, last=(last or []) +[i+1!=len(children)])
        for task in tasklist:
            _display(task)
    else:
        log.info("no tasks to display")


@cli.command()
def docs():
    """
    Opens the Jolt documentation in the default webbrowser.
    """
    webbrowser.open(config.get("jolt", "docs", "http://jolt.readthedocs.io/"))


@cli.command(hidden=True)
@click.argument("task", type=str, nargs=-1, required=True)
@click.option("-r", "--remove", is_flag=True, help="Remove tasks from existing manifest.")
@click.option("-d", "--default", type=str, multiple=True, help="Override default parameter values.")
@click.option("-o", "--output", type=str, default="default.joltxmanifest", help="Manifest filename.")
@click.pass_context
def freeze(ctx, task, default, output, remove):
    """
    Freeze the identity of a task.

    <WIP>
    """
    manifest = ctx.obj["manifest"]

    options = JoltOptions(default=default)
    acache = cache.ArtifactCache.get(options)
    scheduler.ExecutorRegistry.get(options)
    registry = TaskRegistry.get()

    for params in default:
        registry.set_default_parameters(params)

    gb = graph.GraphBuilder(registry, manifest)
    dag = gb.build(task)

    available_in_cache = [
        (t.is_available_locally(acache) or (
            t.is_available_remotely(acache) and acache.download_enabled()), t)
        for t in dag.tasks if t.is_cacheable()]

    for available, task in available_in_cache:
        raise_task_error_if(
            not remove and not available, task,
            "task artifact is not available in any cache, build it first")

    for task in dag.tasks:
        if task.is_resource() or not task.is_cacheable():
            continue
        manifest_task = manifest.find_task(task)
        if remove and manifest_task:
            manifest.remove_task(manifest_task)
            continue
        if not remove:
            if not manifest_task:
                manifest_task = manifest.create_task()
            manifest_task.name = task.qualified_name
            manifest_task.identity = task.identity

    manifest.write(fs.path.join(JoltLoader.get().joltdir, output))


@cli.command(name="list")
@click.argument("task", type=str, nargs=-1, required=False, autocompletion=_autocomplete_tasks)
@click.option("-r", "--reverse", type=str, help="List consumers of REVERSE if TASK is executed.")
@click.pass_context
def _list(ctx, task=None, reverse=None):
    """
    List all tasks, or dependencies of a specific task.

    <WIP>
    """

    raise_error_if(not task and reverse, "TASK required with --reverse")

    registry = TaskRegistry.get()

    if not task:
        classes = registry.get_task_classes()
        classes += registry.get_test_classes()
        for task in sorted(classes, key=lambda x: x.name):
            if task.name:
                print(task.name)
        return

    task = [utils.stable_task_name(t) for t in task]
    reverse = [utils.stable_task_name(t) for t in utils.as_list(reverse or [])]

    try:
        dag = graph.GraphBuilder(registry, ctx.obj["manifest"]).build(task, influence=False)
    except JoltError as e:
        raise e
    except:
        raise_error("an exception occurred during task dependency evaluation, see log for details")

    task = reverse or task
    nodes = dag.select(lambda graph, node: \
                       node.short_qualified_name in task or \
                       node.qualified_name in task)

    tasklist = set()
    iterator = dag.predecessors if reverse else dag.successors

    for node in nodes:
        for task in iterator(node):
            tasklist.add(task.short_qualified_name)

    for task in sorted(list(tasklist)):
        print(task)


@cli.command(name="log")
@click.option("-f", "--follow", is_flag=True, help="Display log output as it appears")
@click.option("-d", "--delete", is_flag=True, help="Delete the log file")
def _log(follow, delete):
    """
    Access the Jolt log file.

    <WIP>
    """
    if follow:
        subprocess.call("tail -f {0}".format(logfile), shell=True)
    elif delete:
        fs.unlink(logfile)
    else:
        t = tools.Tools()
        configured_pager = config.get("jolt", "pager", environ.get("PAGER", None))
        for pager in [configured_pager, "less", "more", "cat"]:
            if pager and t.which(pager):
                return subprocess.call("{1} {0}".format(logfile, pager), shell=True)
        print(t.read_file(logfile))

@cli.command()
@click.argument("task", autocompletion=_autocomplete_tasks)
@click.option("-i", "--influence", is_flag=True, help="Print task influence.")
@click.option("-a", "--artifacts", is_flag=True, help="Print task artifact status.")
@click.option("-s", "--salt", type=str, help="Add salt as task influence.")
@click.pass_context
def info(ctx, task, influence=False, artifacts=False, salt=None):
    """
    View information about a task, including its documentation.

    <WIP>
    """
    task_name = task
    task_cls_name, task_params = utils.parse_task_name(task_name)
    task_registry = TaskRegistry.get()
    task = task_registry.get_task_class(task_cls_name) or \
           task_registry.get_test_class(task_cls_name)
    raise_task_error_if(not task, task_name, "no such task")

    click.echo()
    click.echo("  {0}".format(task.name))
    click.echo()
    if task.__doc__:
        click.echo("  {0}".format(task.__doc__.strip()))
        click.echo()
    click.echo("  Parameters")
    has_param = False
    params = { key: getattr(task, key) for key in dir(task)
               if isinstance(utils.getattr_safe(task, key), Parameter) }
    for item, param in params.items():
        has_param = True
        click.echo("    {0:<15}   {1}".format(item, param.__doc__ or ""))
    if not has_param:
        click.echo("    None")

    click.echo()
    click.echo("  Requirements")
    manifest = ctx.obj["manifest"]
    try:
        task = task_registry.get_task(task_name, manifest=manifest)
        for req in sorted(utils.as_list(utils.call_or_return(task, task.requires))):
            click.echo("    {0}".format(task.tools.expand(req)))
        if not task.requires:
            click.echo("    None")
        click.echo()
    except Exception as e:
        log.exception()
        if "has not been set" in str(e):
            click.echo("    Unavailable (parameters must be set)")
            click.echo()
            return
        click.echo("    Unavailable (exception during evaluation)")
        click.echo()
        return

    if salt:
        task.taint = salt

    if artifacts:
        acache = cache.ArtifactCache.get()
        builder = graph.GraphBuilder(task_registry, manifest)
        dag = builder.build([task.qualified_name])
        tasks = dag.select(lambda graph, node: node.task is task)
        assert len(tasks) == 1, "graph produced multiple tasks, one expected"
        proxy = tasks[0]
        task = proxy.task

        click.echo("  Cache")
        click.echo("    Identity          {0}".format(proxy.identity))
        if acache.is_available_locally(proxy):
            with acache.get_artifact(proxy) as artifact:
                click.echo("    Location          {0}".format(artifact.path))
            click.echo("    Local             True ({0})".format(
                utils.as_human_size(acache.get_artifact(proxy).get_size())))
        else:
            click.echo("    Local             False")
        click.echo("    Remote            {0}".format(acache.is_available_remotely(proxy)))
        click.echo()

    if influence:
        click.echo("  Influence")
        for string in HashInfluenceRegistry.get().get_strings(task):
            string = string.split(":", 1)
            click.echo("    {:<18}{}".format(string[0][10:], string[1].strip()))



import glob
import os
import yaml

from jolt.tasks import Parameter, Task, TaskGenerator, TaskRegistry
from jolt import log
from jolt import loader
from jolt.error import raise_error_if
from jolt import influence


log.verbose("[YamlTask] Loaded")


@influence.attribute("yaml")
class YamlTask(Task):
    name = None
    requires = []
    cacheable = True
    commands = []
    collects = []
    extends = ""
    fast = False
    influence = []
    joltdir = "."
    joltproject = None
    metadata = {}
    selfsustained = False
    taint = False
    weight = False
    yaml = None

    def run(self, d, t):
        self.builddir = t.builddir()
        with t.cwd(self.builddir):
            for cmd in self.commands:
                t.run(cmd)

    def publish(self, a, t):
        with t.cwd(self.builddir):
            for pf in self.collects:
                a.collect(**pf)
            for section, item in self.metadata.items():
                for key, value in item.items():
                    setattr(getattr(a, section), key, value)


class YamlTaskBuilder(object):
    def build(self, yobj):
        return None


_builders = {}
def register(cls):
    raise_error_if(not issubclass(cls, YamlTaskBuilder),
                   "{} is not a YamlTaskBuilder", cls.__name__)
    _builders[cls.name] = cls()


@register
class RegularYamlTaskBuilder(YamlTaskBuilder):
    name = "task"

    def build(self, ytask):
        class LoadedYamlTask(YamlTask):
            name = None
            requires = []
            cacheable = True
            commands = []
            collects = []
            cxxinfo = []
            exports = []
            extends = ""
            fast = False
            influence = []
            joltdir = "."
            joltproject = None
            selfsustained = False
            taint = False
            weight = False
            yaml = None

        for key, obj in ytask.items():
            setattr(LoadedYamlTask, key, obj)

        LoadedYamlTask.yaml = ytask

        return LoadedYamlTask


class YamlRecipe(loader.Recipe):
    def __init__(self, *args, **kwargs):
        super(YamlRecipe, self).__init__(*args, **kwargs)
        self._preload()

    def _preload(self):
        with open(self.path) as f:
            root = yaml.safe_load_all(f)
            root = [o for o in root]

        for doc in root:
            for key, obj in doc.items():
                builder = _builders.get(key)
                if not builder:
                    log.verbose("Unknown yaml task type: {}", key)
                    continue
                task = builder.build(obj)
                task.joltdir = os.path.dirname(self.path)
                self.tasks.append(task)

    def is_valid(self):
        return len(self.tasks) > 0


class YamlLoader(loader.Loader):
    def __init__(self):
        self._recipes = []
        self._find_files()

    def _find_files(self):
        files = []
        oldpath = None
        curpath = os.getcwd()
        while not self._recipes and oldpath != curpath:
            files = glob.glob(os.path.join(curpath, "*.yaml"))

            for filepath in files:
                recipe = YamlRecipe(filepath)
                if recipe.is_valid():
                    self._recipes.append(recipe)

            oldpath = curpath
            curpath = os.path.dirname(oldpath)
        self.path = oldpath

    @property
    def recipes(self):
        return self._recipes


@loader.register
class YamlLoaderFactory(loader.LoaderFactory):
    def create(self):
        return YamlLoader()



from datetime import datetime, timedelta

from jolt import utils


class ArtifactEvictionStrategy(object):
    name = None

    def __init__(self, *args, **kwargs):
        pass

    def is_evictable(self, artifact):
        """ Return True if eviction is permissible. """
        return True

    @property
    def value(self):
        return self.name


class Immediately(ArtifactEvictionStrategy):
    name = "immediately"

    def is_evictable(self, artifact):
        return True


class Never(ArtifactEvictionStrategy):
    name = "never"

    def is_evictable(self, artifact):
        return False


class After(ArtifactEvictionStrategy):
    name = "after"

    def __init__(self, *args, **kwargs):
        self.delta = timedelta(*args, **kwargs)

    def is_evictable(self, artifact):
        return datetime.now() - artifact["created"] > self.delta

    @property
    def value(self):
        return {
            self.name: {
                "days": self.delta.days,
                "seconds": self.delta.seconds
            }
        }


class WhenUnusedFor(ArtifactEvictionStrategy):
    name = "when_unused_for"

    def __init__(self, *args, **kwargs):
        self.delta = timedelta(*args, **kwargs)

    def is_evictable(self, artifact):
        return datetime.now() - artifact["used"] > self.delta

    @property
    def value(self):
        return {
            self.name: {
                "days": self.delta.days,
                "seconds": self.delta.seconds
            }
        }


@utils.Singleton
class ArtifactEvictionStrategyRegister(object):
    def __init__(self):
        self.strategies = {}

    def add(self, strategy):
        self.strategies[strategy.name] = strategy

    def find(self, expiration_data):
        if type(expiration_data) == str:
            return self.strategies.get(expiration_data, Immediately)()
        if type(expiration_data) == dict:
            for strategy, data in expiration_data.items():
                strategy = self.strategies.get(strategy)
                return strategy(**data) if strategy else Immediately()
        return Immediately()


ArtifactEvictionStrategyRegister.get().add(Immediately)
ArtifactEvictionStrategyRegister.get().add(Never)
ArtifactEvictionStrategyRegister.get().add(After)
ArtifactEvictionStrategyRegister.get().add(WhenUnusedFor)



from jolt import config
from jolt import log
from jolt import tasks


log.verbose("[Alias] Loaded")


_registry = tasks.TaskRegistry.get()

for key, value in config.options("alias"):
    class ConfigAlias(tasks.Alias):
        name = key
        requires = value.split()
    _registry.add_task_class(ConfigAlias)



from xml.etree import ElementTree as ET


def _tcname(suite, case):
    return "{}.{}".format(suite.get("name"), case.get("name"))


def import_failures(xml, report):
    tree = ET.parse(xml)
    root = tree.getroot()
    testsuites = root.findall(".//testsuite")
    for testsuite in testsuites:
        testcases = testsuite.findall(".//testcase")
        for testcase in testcases:
            errors = testcase.findall(".//error")
            for error in errors:
                name = _tcname(testsuite, testcase)
                message = error.text
                report.add_error("Test Error", name, message)

            failures = testcase.findall(".//failure")
            for failure in failures:
                name = _tcname(testsuite, testcase)
                message = failure.text
                report.add_error("Test Failed", name, message)



from contextlib import contextmanager, ExitStack
import functools

from jolt import utils


class TaskHook(object):
    def task_created(self, task):
        pass

    def task_started(self, task):
        pass

    def task_finished(self, task):
        pass

    def task_failed(self, task):
        pass

    def task_pruned(self, task):
        pass

    def task_skipped(self, task):
        pass

    def task_prerun(self, task, deps, tools):
        pass

    def task_prepublish(self, task, artifact, tools):
        pass

    def task_prenunpack(self, task, artifact, tools):
        pass

    def task_postrun(self, task, deps, tools):
        pass

    def task_postpublish(self, task, artifact, tools):
        pass

    def task_postunpack(self, task, artifact, tools):
        pass

    @contextmanager
    def task_run(self, task):
        yield


class TaskHookFactory(object):
    @staticmethod
    def register(cls):
        TaskHookRegistry.factories.append((cls, 0))
        TaskHookRegistry.factories.sort(key=lambda x: x[1])

    @staticmethod
    def register_with_prio(prio):
        def decorator(cls):
            TaskHookRegistry.factories.append((cls, prio))
            TaskHookRegistry.factories.sort(key=lambda x: x[1])
        return decorator

    def create(self, env):
        raise NotImplementedError()


@utils.Singleton
class TaskHookRegistry(object):
    factories = []

    def __init__(self, env=None):
        self.env = env
        self.hooks = [factory().create(env) for factory, _ in TaskHookRegistry.factories]
        self.hooks = list(filter(lambda n: n, self.hooks))

    def task_created(self, task):
        for ext in self.hooks:
            utils.call_and_catch_and_log(ext.task_created, task)

    def task_started(self, task):
        if task.is_resource():
            return
        for ext in self.hooks:
            utils.call_and_catch_and_log(ext.task_started, task)

    def task_finished(self, task):
        if task.is_resource():
            return
        for ext in self.hooks:
            utils.call_and_catch_and_log(ext.task_finished, task)

    def task_failed(self, task):
        if task.is_resource():
            return
        for ext in self.hooks:
            utils.call_and_catch_and_log(ext.task_failed, task)

    def task_pruned(self, task):
        if task.is_resource():
            return
        for ext in self.hooks:
            utils.call_and_catch_and_log(ext.task_pruned, task)

    def task_skipped(self, task):
        if task.is_resource():
            return
        for ext in self.hooks:
            utils.call_and_catch_and_log(ext.task_skipped, task)

    def task_prerun(self, task, deps, tools):
        if task.is_resource():
            return
        for ext in self.hooks:
            utils.call_and_catch_and_log(ext.task_prerun, task, deps, tools)

    def task_prepublish(self, task, artifact, tools):
        if task.is_resource():
            return
        for ext in self.hooks:
            utils.call_and_catch_and_log(ext.task_prepublish, task, artifact, tools)

    def task_preunpack(self, task, artifact, tools):
        if task.is_resource():
            return
        for ext in self.hooks:
            utils.call_and_catch_and_log(ext.task_preunpack, task, artifact, tools)

    def task_postrun(self, task, deps, tools):
        if task.is_resource():
            return
        for ext in self.hooks:
            utils.call_and_catch_and_log(ext.task_postrun, task, deps, tools)

    def task_postpublish(self, task, artifact, tools):
        if task.is_resource():
            return
        for ext in self.hooks:
            utils.call_and_catch_and_log(ext.task_postpublish, task, artifact, tools)

    def task_postunpack(self, task, artifact, tools):
        if task.is_resource():
            return
        for ext in self.hooks:
            utils.call_and_catch_and_log(ext.task_postunpack, task, artifact, tools)

    @contextmanager
    def task_run(self, task):
        tasks = [task] if type(task) != list else task
        with ExitStack() as stack:
            for task in tasks:
                if task.is_resource():
                    continue
                for ext in self.hooks:
                    stack.enter_context(ext.task_run(task))
            yield


class CliHook(object):
    @contextmanager
    def cli_build(self, *args, **kwargs):
        yield


class CliHookFactory(object):
    @staticmethod
    def register(cls):
        CliHookRegistry.factories.append(cls)

    def create(self, env):
        raise NotImplementedError()


@utils.Singleton
class CliHookRegistry(object):
    factories = []

    def __init__(self, env=None):
        self.env = env
        self.hooks = [factory().create(env) for factory in CliHookRegistry.factories]
        self.hooks = list(filter(lambda n: n, self.hooks))

    @contextmanager
    def cli_build(self, *args, **kwargs):
        with ExitStack() as stack:
            for ext in self.hooks:
                stack.enter_context(ext.cli_build(*args, **kwargs))
            yield


def task_created(task):
    TaskHookRegistry.get().task_created(task)

def task_started(task):
    TaskHookRegistry.get().task_started(task)

def task_failed(task):
    TaskHookRegistry.get().task_failed(task)

def task_finished(task):
    TaskHookRegistry.get().task_finished(task)

def task_pruned(task):
    TaskHookRegistry.get().task_pruned(task)

def task_skipped(task):
    TaskHookRegistry.get().task_skipped(task)

def task_prerun(task, deps, tools):
    TaskHookRegistry.get().task_prerun(task, deps, tools)

def task_prepublish(task, artifact, tools):
    TaskHookRegistry.get().task_prepublish(task, artifact, tools)

def task_preunpack(task, artifact, tools):
    TaskHookRegistry.get().task_preunpack(task, artifact, tools)

def task_postrun(task, deps, tools):
    TaskHookRegistry.get().task_postrun(task, deps, tools)

def task_postpublish(task, artifact, tools):
    TaskHookRegistry.get().task_postpublish(task, artifact, tools)

def task_postunpack(task, artifact, tools):
    TaskHookRegistry.get().task_postunpack(task, artifact, tools)

def task_run(task):
    return TaskHookRegistry.get().task_run(task)

def cli_build(cmd):
    @functools.wraps(cmd)
    def decorator(*args, **kwargs):
        with CliHookRegistry.get().cli_build():
            return cmd(*args, **kwargs)
    return decorator



from setuptools import setup, find_packages
from setuptools.command.build_py import build_py
from codecs import open
from os import makedirs, path
from shutil import copyfile

here = path.abspath(path.dirname(__file__))
name = "jolt"
version = "0.9.22"


# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


class BuildCommand(build_py):
    def run(self):
        build_py.run(self)

        # Install additional files required by selfdeploy plugin
        if not self.dry_run:
            target_dir = path.join(self.build_lib, name, "plugins", "selfdeploy")
            makedirs(target_dir)
            for fn in ["setup.py", "README.rst"]:
                copyfile(path.join(here, fn), path.join(target_dir,fn))

setup(
    name=name,
    cmdclass={"build_py": BuildCommand},
    version=version,
    description='A task executor',
    long_description=long_description,
    url='https://bitbucket.org/rand_r/jolt',
    author='Robert Andersson',
    author_email='rand_r@bitbucket.org',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Programming Language :: Python :: 3',
    ],
    keywords='build msbuild xcode make pam bazel jolt conan',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=["requests", "click>=7.0", "keyring", "keyrings.alt",
                      "jinja2", "python-jenkins", "multi_key_dict",
                      "tqdm", "bz2file", "colorama", "ninja-syntax",
                      "pyyaml", "fasteners", "pygit2", "lxml"],
    dependency_links=[],
    extras_require={
        'allure': ['allure-python-commons'],
        'amqp': ['pika'],
        'dev': ['check-manifest'],
        'test': ['coverage'],
    },
    package_data={
        'jolt': ['**/*.xslt'],
    },
    entry_points={
        'console_scripts': [
            'jolt=jolt.__main__:main',
        ],
    },
)



from requests.auth import HTTPBasicAuth
from requests.exceptions import ConnectTimeout, RequestException
import keyring
import getpass


from jolt import utils
from jolt import cache
from jolt import log
from jolt import config
from jolt import filesystem as fs
from jolt.error import raise_error_if, JoltError


NAME = "http"
TIMEOUT = (3.5, 27)
TIMEOUT_HEAD = (27, 27)


class Http(cache.StorageProvider):
    def __init__(self, cache):
        super(Http, self).__init__()
        self._cache = cache
        self._uri = config.get(NAME, "uri")
        raise_error_if(not self._uri, "HTTP URI not configured")
        if self._uri[-1] != "/":
            self._uri += "/"
        self._upload = config.getboolean(NAME, "upload", True)
        self._download = config.getboolean(NAME, "download", True)
        self._disabled = False

    def _get_auth(self):
        service = config.get(NAME, "keyring.service")
        if not service:
            return None

        username = config.get(NAME, "keyring.username")
        if not username:
            username = utils.read_input(NAME + " username: ")
            raise_error_if(not username, "no username configured for " + NAME)
            config.set(NAME, "keyring.username", username)
            config.save()

        password = config.get(NAME, "keyring.password") or \
                   keyring.get_password(NAME, username)
        if not password:
            password = getpass.getpass(NAME + " password: ")
            raise_error_if(not password, "no password in keyring for " + NAME)
            keyring.set_password(service, username, password)

        return HTTPBasicAuth(username, password)

    def _get_url(self, node, artifact):
        return "{uri}/{name}/{file}".format(
            uri=self._uri,
            name=node.name,
            file=fs.path.basename(artifact.get_archive_path()))

    @utils.retried.on_exception((RequestException, JoltError))
    def download(self, node, force=False):
        if self._disabled:
            return False
        if not self._download and not force:
            return False
        with self._cache.get_artifact(node) as artifact:
            url = self._get_url(node, artifact)
            if node.tools.download(url, artifact.get_archive_path(), exceptions=True, timeout=TIMEOUT):
                return True
        return False

    def download_enabled(self):
        return not self._disabled and self._download

    @utils.retried.on_exception((RequestException))
    def upload(self, node, force=False):
        if self._disabled:
            return True
        if not self._upload and not force:
            return True
        with self._cache.get_artifact(node) as artifact:
            url = self._get_url(node, artifact)
            archive = artifact.get_archive()
            return node.tools.upload(archive, url, exceptions=True,
                                     auth=self._get_auth(), timeout=TIMEOUT)
        return False

    def upload_enabled(self):
        return not self._disabled and self._upload

    @utils.retried.on_exception((RequestException))
    def location(self, node):
        if self._disabled:
            return False
        with self._cache.get_artifact(node) as artifact:
            from requests.api import head

            url = self._get_url(node, artifact)
            try:
                response = head(url, stream=True, timeout=TIMEOUT_HEAD)
            except ConnectTimeout:
                self._disabled = True
                log.warning("[HTTP] failed to establish server connection, disabled")
                return False

            log.debug("[HTTP] Head: {0}", url)
            log.debug("[HTTP] Response: {0}", response.status_code)
            return url if response.status_code == 200 else ''
        return False


@cache.RegisterStorage
class HttpFactory(cache.StorageProviderFactory):
    @staticmethod
    def create(cache):
        log.verbose("[Http] Loaded")
        return Http(cache)



import glob
import sys
import os
import subprocess
import hashlib
import json
import errno
import uuid
join = os.path.join


_hash_cache = {}


class Depfile(object):
    def __init__(self, objfile, dependencies=None, cmdline=None):
        self.valid = False
        self.product = ""
        self.cmdline = cmdline
        self.path = objfile + ".d"
        self.dependencies = []

        if not dependencies:
            if not os.path.exists(self.path):
                return

            with open(self.path) as f:
                data = f.read()

            data = data.split(":", 1)
            if len(data) <= 1:
                return

            dependencies = data[1]
            dependencies = dependencies.split()
            dependencies = [f.rstrip("\\").strip() for f in dependencies]
            dependencies = [os.path.normpath(f)
                            for f in filter(lambda n: n, dependencies)]

        self.product = objfile
        self.dependencies = dependencies or []
        self.valid = True

    def _hash_file(self, filepath):
        digest = _hash_cache.get(filepath)
        if digest:
            return digest

        sha = hashlib.sha1()
        try:
            with open(filepath, "rb") as f:
                sha.update(f.read())
        except:
            sha.update(str(uuid.uuid4()).encode())

        _hash_cache[filepath] = digest = sha.hexdigest()
        return digest

    @property
    def hash_cmdline(self):
        sha_cmd = hashlib.sha1(str(self.cmdline).replace(os.getenv("JOLT_CACHEDIR"), "").encode())
        return sha_cmd.hexdigest()

    @property
    def hash_deps(self):
        if type(self.dependencies) == str:
            return self.dependencies
        sha = hashlib.sha1()
        for dep in sorted(self.dependencies):
            fh = self._hash_file(dep)
            sha.update(fh.encode())
        return sha.hexdigest()

    @property
    def hash(self):
        sha = hashlib.sha1()
        sha.update(self.hash_cmdline.encode())
        sha.update(self.hash_deps.encode())
        return sha.hexdigest()


class LibraryManifest(object):
    def __init__(self, path):
        self._path = path
        self._data = {}
        self._objects = set()

    def read(self):
        try:
            with open(self._path, "r") as f:
                self._data = json.loads(f.read())
        except:
            pass

    def write(self):
        with open(self._path, "w") as f:
            f.write(json.dumps(self._data, indent=2))

    def add_library(self, library_file):
        self._data["filename"] = library_file

    def add_object_file(self, objpath, cmdline):
        depfile = Depfile(objpath, cmdline=cmdline)
        objects = self.objects
        objects[objpath] = {"digest": depfile.hash, "deps": depfile.dependencies}
        self._data["objects"] = objects

        if objpath in self._objects:
            print("error: object file name collision: {}".format(objpath))
            sys.exit(1)
        self._objects.add(objpath)

    def get_object(self, objpath, builddir):
        objdir = join(builddir, os.path.dirname(objpath))
        data = self.objects.get(objpath)
        if not data:
            return False

        try:
            os.makedirs(objdir)
        except OSError as e:
            if e.errno != errno.EEXIST:
                return False

        rv = subprocess.call([os.getenv("AR", "ar"), "x", self.library, objpath], cwd=objdir)
        if rv != 0:
            return False

        with open(objpath+".d", "w") as f:
            f.write("{}: {}\n".format(objpath, " ".join(data["deps"])))
            return True

        return False


    @property
    def objects(self):
        return self._data.get("objects", {})

    @property
    def library(self):
        return os.path.join(os.path.dirname(self._path), self._data["filename"])

    @property
    def path(self):
        return self._path


class Cache(object):
    def __init__(self, builddir):
        self._builddir = builddir
        self._objects = {}
        self._objnames = {}
        self._sources = {}

    def add_manifest(self, manifest):
        for objpath, objdata in manifest.objects.items():
            objname = os.path.basename(objpath)
            objpath_known = self._objnames.get(objname, objpath)
            if objpath_known != objpath:
                # Multiple objects with the same basename.
                # No way to know which object will be extracted
                # from the archive. Remove entries and fall back
                # to compilation.
                try:
                    self._objnames[objname] = None
                    del self._objects[objpath_known]
                except KeyError:
                    pass
                continue
            self._objnames[objname] = objpath
            obj = self._objects.get(str(objpath), [])
            deps = [join(self._builddir, dep) for dep in objdata["deps"]]
            obj.append({
                "digest": objdata["digest"],
                "manifest": manifest.path,
                "deps": Depfile(objpath, dependencies=deps).hash_deps})
            self._objects[str(objpath)] = obj

    def load_manifests(self, cachedir, task):
        pattern = "{cachedir}/{task}/*/.ninja.json".format(cachedir=cachedir, task=task)
        manifests = glob.glob(pattern)

        maxartifacts = int(os.getenv("NINJACACHE_MAXARTIFACTS", 0))
        if maxartifacts > 0:
            manifests = manifests[:maxartifacts]

        for manifest in manifests:
            manifest = LibraryManifest(manifest)
            manifest.read()
            self.add_manifest(manifest)

    def load(self):
        with open(join(self._builddir, ".cache.json")) as f:
            self._objects = json.load(f)

    def save(self):
        with open(join(self._builddir, ".cache.json"), "w") as f:
            json.dump(self._objects, f, indent=2)

    def lookup(self, objpath, cmdline):
        for objdata in self._objects.get(objpath, []):
            dep = Depfile(objpath, objdata["deps"], cmdline)
            if dep.hash == objdata["digest"]:
                manifest = LibraryManifest(objdata["manifest"])
                manifest.read()
                return manifest


class LockFile(object):
    def __init__(self, path=".", *args, **kwargs):
        from fasteners import process_lock
        self._file = process_lock.InterProcessLock(os.path.join(path, ".ninja.lock"))

    def __enter__(self, *args, **kwargs):
        self._file.acquire()
        return self

    def __exit__(self, *args, **kwargs):
        self._file.release()
        pass


###############################################################################

def argscan(args, arg):
    try:
        index = args.index(arg)
    except ValueError:
        return None
    if len(args) <= index+1:
        return None
    return args[index+1]


def verbose(fmt, *args, **kwargs):
    if os.getenv("NINJACACHE_VERBOSE", "0") == "1":
        print(fmt.format(*args, **kwargs))


def cli(compiler_args):
    objfile = argscan(sys.argv, "-o")

    if os.getenv("NINJACACHE_DISABLE", "0") != "1":
        cache = Cache(os.getcwd())
        cache.load()
        manifest = cache.lookup(objfile, compiler_args)
        if manifest:
            if manifest.get_object(objfile, os.getcwd()):
                verbose("Reusing {} from {}", objfile, manifest.library)
                with LockFile():
                    manifest = LibraryManifest(".ninja.json")
                    manifest.read()
                    manifest.add_object_file(objfile, compiler_args)
                    manifest.write()
                    sys.exit(0)
            else:
                verbose("Extraction from library failed, rule will be executed normally")

    rv = subprocess.call(compiler_args)
    if rv != 0:
        sys.exit(rv)

    with LockFile():
        manifest = LibraryManifest(".ninja.json")
        manifest.read()
        manifest.add_object_file(objfile, compiler_args)
        manifest.write()


if __name__ == "__main__":
    compiler_args = sys.argv[sys.argv.index("--")+1:]
    cli(compiler_args)
